/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "filelocator.h"
#include "font.h"
#include "state.h"
#include "texture.h"
#include "transition.h"

// Bits zum Speichern der Tastenzustnde
enum KeyFlag
{
	KF_DOWN = 1,			// Taste ist gerade gedrckt
	KF_PRESSED = 2,			// Taste wurde seit dem letzten Update gedrckt
	KF_REALLY_PRESSED = 4,	// Taste wurde seit dem letzten Update tatschlich gedrckt
	KF_RELEASED = 8			// Taste wurde seit dem letzten Update losgelassen
};

Application::Application(const Vec2i& requestedScreenResolution,
						 bool fullScreen,
						 float updateRate,
						 const std::string& windowTitle,
						 const std::string& windowIconFilename)
	: doQuit(false)
	, active(true)
	, screenResolution(0, 0)
	, fullScreen(fullScreen)
	, viewportAspect(0.0f)
	, updateRate(updateRate)
	, updateStepSize(1.0f / updateRate)
	, time(0.0)
	, p_fileLocator(0)
	, p_textureManager(0)
	, p_soundManager(0)
	, p_fontManager(0)
	, p_transition(0)
{
	assert(updateRate > 0.0f);
	transitionTextureIDs[0] = 0;
	transitionTextureIDs[1] = 0;

	// SDL initialisieren
	SDL_Init(SDL_INIT_TIMER | SDL_INIT_VIDEO);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	SDL_EnableUNICODE(1);

	// Fenstertitel setzen
	if(!windowTitle.empty()) SDL_WM_SetCaption(windowTitle.c_str(), 0);

	if(!windowIconFilename.empty())
	{
		// Fenster-Icon setzen
		SDL_Surface* p_surface = IMG_Load(windowIconFilename.c_str());
		if(p_surface)
		{
			SDL_WM_SetIcon(p_surface, 0);
			SDL_FreeSurface(p_surface);
		}
	}

	// OpenGL initialisieren
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 1);
	Uint32 flags = SDL_OPENGL;
	if(fullScreen) flags |= SDL_FULLSCREEN;
	SDL_SetVideoMode(requestedScreenResolution.x, requestedScreenResolution.y, 32, flags);

	// echte Bildschirmgre abfragen
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	screenResolution = Vec2i(viewport[2], viewport[3]);
	setViewport(Vec2i(0, 0), screenResolution);

	// OpenAL initialisieren
	p_openALDevice = alcOpenDevice(0);
	p_openALContext = alcCreateContext(p_openALDevice, 0);
	alcMakeContextCurrent(p_openALContext);

	// Tastenzustnde zurcksetzen
	for(int i = 0; i < SDLK_LAST; ++i) keyState[i] = 0;

	// FileLocator erzeugen
	p_fileLocator = new FileLocator;

	// Texturmanager initialisieren
	p_textureManager = new TextureManager(*this, *p_fileLocator);

	// Soundmanager initialisieren
	p_soundManager = new SoundManager(*this, *p_fileLocator);
	
	// Font-Manager initialisieren
	p_fontManager = new FontManager(*this, *p_fileLocator);

	// Texturen fr bergangseffekte erzeugen
	glGenTextures(2, transitionTextureIDs);
	for(int i = 0; i < 2; ++i)
	{
		assert(transitionTextureIDs[i]);
		transitionTextureSize = Vec2i(nextPow2(screenResolution.x), nextPow2(screenResolution.y));
		glBindTexture(GL_TEXTURE_2D, transitionTextureIDs[i]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, transitionTextureSize.x, transitionTextureSize.y, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);
	}
}

Application::~Application()
{
	// Screenshot-Sound freigeben
	screenshotSound.reset();

	// Zustnde lschen
	for(StateMapType::const_iterator it = stateRegistry.begin();
		it != stateRegistry.end();
		++it)
	{
		delete it->second;
	}

	// FileLocator lschen
	delete p_fileLocator;

	// Texturmanager lschen
	delete p_textureManager;

	// Soundmanager lschen
	delete p_soundManager;

	// OpenAL herunterfahren
	alcDestroyContext(p_openALContext);
	alcCloseDevice(p_openALDevice);

	// bergangseffekt und Texturen lschen
	delete p_transition;
	glDeleteTextures(2, transitionTextureIDs);

	// SDL herunterfahren
	SDL_Quit();
}

void Application::run()
{
	float timeToProcess = 0.0f;
	bool needToRender = false;
	bool needToSwap = false;
	Uint32 startTime = SDL_GetTicks();

	while(!doQuit)
	{
		const Uint32 now = SDL_GetTicks();
		const float dt = 0.001f * (now - startTime);
		startTime = now;

		// Die vergangene Zeit muss verarbeitet werden.
		timeToProcess += dt;

		// zu verarbeitende Zeit begrenzen
		timeToProcess = std::min(timeToProcess, 4.0f * updateStepSize);

		if(active && needToRender)
		{
			// rendern
			render();
			needToRender = false;
			needToSwap = true;
		}

		while(timeToProcess >= updateStepSize)
		{
			// aktualisieren
			update();
			if(doQuit) break;
			timeToProcess -= updateStepSize;
			needToRender = true;
		}

		if(needToSwap)
		{
			if(p_transition)
			{
				if(p_transition->getNeedsNewImageTexture())
				{
					// aktuelles Bild in der Textur fr das neue Bild speichern
					glBindTexture(GL_TEXTURE_2D, transitionTextureIDs[1]);
					glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, transitionTextureSize.y - screenResolution.y, 0, 0, screenResolution.x, screenResolution.y);
				}

				// Texturmatrix setzen, die anderen Matrizen zurcksetzen
				glMatrixMode(GL_TEXTURE);
				glLoadIdentity();
				glScalef(1.0f / transitionTextureSize.x, -1.0f / transitionTextureSize.y, 1.0f);
				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();

				// Texturierung und Blending aktivieren, Farbe auf Wei setzen
				setTexturing(true);
				setBlending(true);
				glColor4fv(RGBA::WHITE);

				// bergangseffekt rendern
				p_transition->render(transitionTextureIDs[0],
									 p_transition->getNeedsNewImageTexture() ? transitionTextureIDs[1] : 0);
			}

			// anzeigen
			SDL_GL_SwapBuffers();
			needToSwap = false;
		}

		if(!getTopState())
		{
			// kein Zustand mehr - beenden
			doQuit = true;
		}

		// Ist noch genug Zeit brig fr ein kleines Schlfchen?
		const double timeLeft = updateStepSize - timeToProcess;
		if(timeLeft >= 0.001f)
		{
			// ja - kurz schlafen
			SDL_Delay(1);
		}

		// Wenn die Anwendung inaktiv ist, knnen wir ein bisschen schlafen.
		if(!active) SDL_Delay(250);
	}

	// alle restlichen Zustnde verlassen
	while(getTopState()) popState();
}

void Application::registerState(State* p_state)
{
	assert(p_state);
	stateRegistry[p_state->getName()] = p_state;
}

State* Application::getStateByName(const std::string& stateName) const
{
	if(stateName.empty()) return 0;
	else
	{
		StateMapType::const_iterator it = stateRegistry.find(stateName);
		assert(it != stateRegistry.end());
		return it->second;
	}
}

void Application::pushStateByName(const std::string& stateName,
								  const ParameterMap& params)
{
	// alten Zustand holen und deaktivieren
	if(State* p_oldState = getTopState()) p_oldState->onLostFocus();

	// neuen Zustand holen und betreten, dann aktivieren
	State* p_newState = getStateByName(stateName);
	assert(p_newState);
	p_newState->onEnter(params);
	p_newState->onGotFocus();
	states.push(p_newState);
}

void Application::popState()
{
	assert(!states.empty());

	// aktuellen Zustand holen, deaktivieren und verlassen
	State* p_oldState = getTopState();
	p_oldState->onLostFocus();
	p_oldState->onLeave();
	states.pop();

	// neuen Zustand holen und aktivieren
	if(State* p_newState = getTopState()) p_newState->onGotFocus();
}

State* Application::getTopState() const
{
	return states.empty() ? 0 : states.top();
}

void Application::quit()
{
	doQuit = true;
}

void Application::setMusic(const std::string& filename,
						   float volume)
{
	// Wenn die Musik schon gespielt wird, muss nichts getan werden
	// (nur die Lautstrke muss evtl. angepasst werden)
	if(currentMusicFilename == filename)
	{
		currentMusicInst->fade(volume, 1.0f, 1.0f, 2.0f);
		return;
	}

	if(currentMusicInst && currentMusicInst->isValid())
	{
		// aktuelle Musik ausblenden und dann stoppen
		currentMusicInst->fade(0.0f, 1.0f, 1.0f, 2.0f, SoundInstance::FAE_STOP);
		oldMusic = currentMusic;
	}

	currentMusic.reset();
	currentMusicInst.reset();

	if(!filename.empty())
	{
		// neue Musik laden und einblenden
		ParameterMap params;
		params.set<bool>("streaming", true);
		currentMusic = p_soundManager->getResourceHandle(filename, params);
		currentMusicInst = currentMusic->createInstance(true, true, 0.0f);
		currentMusicInst->fade(volume, 1.0f, 1.0f, 2.0f);
	}

	currentMusicFilename = filename;
}

void Application::setScreenshotSound(SoundHandle screenshotSound)
{
	this->screenshotSound = screenshotSound;
}

void Application::startTransition(Transition* p_transition)
{
	// aktuelles Bild in der Textur fr das alte Bild speichern
	glBindTexture(GL_TEXTURE_2D, transitionTextureIDs[0]);
	glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, transitionTextureSize.y - screenResolution.y, 0, 0, screenResolution.x, screenResolution.y);

	// alten bergang lschen, falls vorhanden
	delete this->p_transition;
	this->p_transition = p_transition;
}

void Application::clearViewport(const RGBA& color)
{
	glClearColor(color.r, color.g, color.b, color.a);
	glPushAttrib(GL_SCISSOR_BIT);
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	glScissor(viewport[0], viewport[1], viewport[2], viewport[3]);
	glEnable(GL_SCISSOR_TEST);
	glClear(GL_COLOR_BUFFER_BIT);
	glPopAttrib();
}

void Application::setViewport(const Vec2i& topLeft,
							  const Vec2i& size)
{
	// Koordinaten umrechnen
	glViewport(topLeft.x, screenResolution.y - topLeft.y - size.y, size.x, size.y);

	// Seitenverhltnis speichern
	viewportAspect = static_cast<float>(size.x) / size.y;
}

void Application::setCursor(bool enable)
{
	SDL_ShowCursor(enable ? SDL_ENABLE : SDL_DISABLE);
}

void Application::setTexturing(bool enable)
{
	if(enable) glEnable(GL_TEXTURE_2D);
	else glDisable(GL_TEXTURE_2D);
}

void Application::setBlending(bool enable,
								   GLenum srcFactor,
								   GLenum destFactor)
{
	if(enable)
	{
		glEnable(GL_BLEND);
		glBlendFunc(srcFactor, destFactor);

		if(srcFactor == GL_SRC_ALPHA)
		{
			glEnable(GL_ALPHA_TEST);
			glAlphaFunc(GL_GREATER, 0.0f);
		}
		else
		{
			glDisable(GL_ALPHA_TEST);
		}
	}
	else
	{
		glDisable(GL_BLEND);
		glDisable(GL_ALPHA_TEST);
	}
}

bool Application::isKeyDown(SDLKey key) const
{
	return (keyState[key] & KF_DOWN) != 0;
}

bool Application::wasKeyPressed(SDLKey key) const
{
	return (keyState[key] & KF_PRESSED) != 0;
}

bool Application::wasKeyReallyPressed(SDLKey key) const
{
	return (keyState[key] & KF_REALLY_PRESSED) != 0;
}

bool Application::wasKeyReleased(SDLKey key) const
{
	return (keyState[key] & KF_RELEASED) != 0;
}

const Vec2i& Application::getScreenResolution() const
{
	return screenResolution;
}

float Application::getViewportAspect() const
{
	return viewportAspect;
}

bool Application::getFullScreen() const
{
	return fullScreen;
}

float Application::getUpdateRate() const
{
	return updateRate;
}

float Application::getUpdateStepSize() const
{
	return updateStepSize;
}

float Application::getTime() const
{
	return static_cast<float>(time);
}

FileLocator& Application::getFileLocator() const
{
	return *p_fileLocator;
}

TextureManager& Application::getTextureManager() const
{
	return *p_textureManager;
}

SoundManager& Application::getSoundManager() const
{
	return *p_soundManager;
}

FontManager& Application::getFontManager() const
{
	return *p_fontManager;
}

void Application::update()
{
	// Tastenzustnde fr "gedrckt" und "losgelassen" zurcksetzen
	for(int i = 0; i < SDLK_LAST; ++i)
	{
		keyState[i] &= ~KF_PRESSED;
		keyState[i] &= ~KF_REALLY_PRESSED;
		keyState[i] &= ~KF_RELEASED;
	}

	// Key-Event Queue leeren
	while(!keyQueue.empty()) keyQueue.pop();

	// SDL-Events verarbeiten
	SDL_Event evt;
	while(SDL_PollEvent(&evt))
	{
		switch(evt.type)
		{
		case SDL_QUIT:
			doQuit = true;
			return;
		case SDL_KEYDOWN:
			if(!isKeyDown(evt.key.keysym.sym)) keyState[evt.key.keysym.sym] |= KF_REALLY_PRESSED;
			keyState[evt.key.keysym.sym] |= KF_DOWN;
			keyState[evt.key.keysym.sym] |= KF_PRESSED;
			// Es werden nur anzeigbare Tastendrcke gespeichert
			if(evt.key.keysym.unicode) keyQueue.push(evt.key.keysym);
			if(evt.key.keysym.sym == SDLK_PRINT)
			{
				if(screenshotSound) screenshotSound->createInstance();
				saveScreenshot();
			}
			break;
		case SDL_KEYUP:
			keyState[evt.key.keysym.sym] &= ~KF_DOWN;
			keyState[evt.key.keysym.sym] |= KF_RELEASED;
			break;
		case SDL_ACTIVEEVENT:
			if(evt.active.state & SDL_APPACTIVE || evt.active.state & SDL_APPINPUTFOCUS) active = evt.active.gain != 0;
			break;
		}
	}

	// Sounds aktualisieren
	p_soundManager->update();

	if(active)
	{
		if(State* p_state = getTopState())
		{
			// aktuellen Zustand aktualisieren
			p_state->update();
		}

		// bergangseffekt aktualisieren und lschen, wenn er fertig ist
		if(p_transition && !p_transition->update())
		{
			delete p_transition;
			p_transition = 0;
		}

		time += updateStepSize;
	}
}

bool Application::hasMoreKeys() const
{
	return !keyQueue.empty();
}

SDL_keysym Application::popKey()
{
	SDL_keysym key = keyQueue.front();
	keyQueue.pop();
	return key;
}

const Vec2i& Application::getTransitionTextureSize() const
{
	return transitionTextureSize;
}

void Application::render()
{
	if(State* p_state = getTopState())
	{
		// Matrizen zurcksetzen
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		// aktuellen Zustand rendern
		p_state->render();
	}
}

void Application::saveScreenshot()
{
	char* p_pixelsIn = new char[screenResolution.x * screenResolution.y * 3];
	char* p_pixelsOut = new char[screenResolution.x * screenResolution.y * 3];
	glReadBuffer(GL_FRONT);
	glReadPixels(0, 0, screenResolution.x, screenResolution.y, GL_RGB, GL_UNSIGNED_BYTE, p_pixelsIn);
	glReadBuffer(GL_BACK);
	SDL_Surface* p_shot = SDL_CreateRGBSurfaceFrom(p_pixelsOut, screenResolution.x, screenResolution.y, 24, screenResolution.x * 3, 0x000000FF, 0x0000FF00, 0x00FF0000, 0);

	for(int y = 0; y < screenResolution.y; ++y)
	{
		const char* p_rowIn = p_pixelsIn + y * screenResolution.x * 3;
		char* p_rowOut = p_pixelsOut + (screenResolution.y - 1 - y) * screenResolution.x * 3;
		for(int x = 0; x < screenResolution.x * 3; ++x) p_rowOut[x] = p_rowIn[x];
	}

	char filename[256] = "";
	int screenshotNo = 1;
	while(true)
	{
		sprintf(filename, "screenshots/screenshot%04d.bmp", screenshotNo++);
		if(std::ifstream(filename).fail()) break;
	}

	SDL_SaveBMP(p_shot, filename);
	SDL_FreeSurface(p_shot);
	delete[] p_pixelsIn;
	delete[] p_pixelsOut;
}
