/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#ifndef APPLICATION_H_INCLUDED
#define APPLICATION_H_INCLUDED

#include "sound.h"

class FileLocator;
class FontManager;
class SoundManager;
class State;
class TextureManager;
class Transition;

class Application
	: NonCopyable
{
public:
	// Konstruktor: initialisiert die Anwendung und die Bibliotheken
	Application(const Vec2i& requestedScreenResolution, bool fullScreen, float updateRate, const std::string& windowTitle = "", const std::string& windowIconFilename = "");

	// Destruktor: fhrt die Anwendung und die Bibliotheken herunter
	~Application();

	// fhrt die Hauptschleife aus
	void run();

	// liefert den Zustand mit dem angegebenen Namen (muss vorher mit registerState registriert worden sein)
	State* getStateByName(const std::string& stateName) const;

	// registriert einen Zustand, der dann in den Besitz der Anwendung bergeht (sie lscht ihn spter)
	void registerState(State* p_state);

	// legt einen Zustand auf den Stack, betritt ihn und aktiviert ihn
	void pushStateByName(const std::string& stateName, const ParameterMap& params = ParameterMap());

	// nimmt den obersten Zustand vom Stack, deaktiviert ihn und verlsst ihn
	void popState();

	// liefert den obersten Zustand (den "aktuellen" Zustand)
	State* getTopState() const;

	// verlsst die Hauptschleife
	void quit();

	// setzt die globale Hintergrundmusik (leerer String stoppt sie)
	void setMusic(const std::string& filename, float volume = 1.0f);

	// setzt den Sound fr Screenshots
	void setScreenshotSound(SoundHandle screenshotSound);

	// startet einen bergangseffekt (das Anwendungsobjekt bernimmt die Verantwortung dafr)
	void startTransition(Transition* p_transition);

	// leert den Viewport
	void clearViewport(const RGBA& color);

	// setzt den Viewport
	void setViewport(const Vec2i& topLeft, const Vec2i& size);

	// aktiviert/deaktiviert die Texturierung
	void setTexturing(bool enable);

	// richtet Alpha-Blending ein
	void setBlending(bool enable, GLenum srcFactor = GL_SRC_ALPHA, GLenum destFactor = GL_ONE_MINUS_SRC_ALPHA);

	// zeigt oder versteckt den Mauszeiger
	void setCursor(bool enable);

	// testet, ob eine Taste gerade gedrckt ist
	bool isKeyDown(SDLKey key) const;

	// testet, ob eine Taste seit dem letzten Update gedrckt wurde
	bool wasKeyPressed(SDLKey key) const;

	// testet, ob eine Taste seit dem letzten Update tatschlich gedrckt wurde
	bool wasKeyReallyPressed(SDLKey key) const;

	// testet, ob eine Taste seit dem letzten Update losgelassen wurde
	bool wasKeyReleased(SDLKey key) const;

	// liefert die Bildschirmauflsung
	const Vec2i& getScreenResolution() const;

	// liefert das Viewport-Seitenverhltnis (Breite / Hhe)
	float getViewportAspect() const;

	// liefert die Einstellung fr Vollbild-/Fenstermodus
	bool getFullScreen() const;

	// liefert die Aktualisierungsrate (Updates pro Sekunde)
	float getUpdateRate() const;

	// liefert die Lnge eines Aktualisierungsschritts (in Sekunden)
	float getUpdateStepSize() const;

	// liefert die Zeit zurck, die seit dem Start der Anwendung vergangen ist (in Sekunden)
	float getTime() const;

	// liefert den FileLocator
	FileLocator& getFileLocator() const;

	// liefert den Texturmanager
	TextureManager& getTextureManager() const;

	// liefert den Soundmanager
	SoundManager& getSoundManager() const;

	// liefert den Font-Manager
	FontManager& getFontManager() const;

	// sind noch Tastendrcke in der Schlange?
	bool hasMoreKeys() const;

	// ltesten Tastendruck aus der Schlange nehmen
	SDL_keysym popKey();

	// liefert die Gre der bergangstextur
	const Vec2i& getTransitionTextureSize() const;

private:
	typedef std::stack<State*> StateStackType;
	typedef std::map<std::string, State*> StateMapType;

	// die Anwendung um einen Schritt aktualisieren
	void update();

	// alles rendern
	void render();

	// Screenshot machen
	void saveScreenshot();

	char keyState[SDLK_LAST];				// Zustand aller Tasten (Bitfeld)
	ALCdevice* p_openALDevice;				// OpenAL-Gert
	ALCcontext* p_openALContext;			// OpenAL-Kontext
	StateStackType states;					// Stack mit allen aktiven Zustnden
	StateMapType stateRegistry;				// Map aller registrierten Zustnde
	bool doQuit;							// steuert den Austritt aus der Hauptschleife
	bool active;							// Anwendung aktiv?
	Vec2i screenResolution;					// Bildschirmauflsung
	bool fullScreen;						// Vollbild oder Fenstermodus?
	float viewportAspect;					// Seitenverhltnis des Viewports
	const float updateRate;					// Anzahl der Updates pro Sekunde
	const float updateStepSize;				// Lnge eines Update-Schritts (1 / updateRate)
	double time;							// vergangene Zeit seit Start der Anwendung
	FileLocator* p_fileLocator;				// der FileLocator
	TextureManager* p_textureManager;		// der Texturmanager
	SoundManager* p_soundManager;			// der Soundmanager
	FontManager* p_fontManager;				// der Font-Manager
	std::queue<SDL_keysym> keyQueue;		// Speichert anzeigbare Tastendrcke
	SoundHandle currentMusic;				// aktuelle Musik
	std::string currentMusicFilename;		// Dateiname der aktuellen Musik
	SoundInstanceHandle currentMusicInst;	// Instanz der aktuellen Musik
	SoundHandle oldMusic;					// alte Musik (die ausgeblendet wird)
	SoundHandle screenshotSound;			// Sound fr Screenshots
	GLuint transitionTextureIDs[2];			// Texturen fr bergangseffekte
	Vec2i transitionTextureSize;			// Gre der bergangstextur
	Transition* p_transition;				// aktueller bergangseffekt
};

#endif
