/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#ifndef BASICRESOURCEMANAGER_H_INCLUDED
#define BASICRESOURCEMANAGER_H_INCLUDED

#include "filelocator.h"
#include "handle.h"
#include "resourcemanager.h"

// Makro zum Implementieren eines einfachen Ressourcenmanagers
#define BEGIN_IMPLEMENT_BASIC_RESOURCE_MANAGER(MANAGER_TYPE, RESOURCE_TYPE) \
	class MANAGER_TYPE \
		: public BasicResourceManager<RESOURCE_TYPE, MANAGER_TYPE> \
	{ \
	public: \
		MANAGER_TYPE(Application& app, FileLocator& fileLocator) \
			: BasicResourceManager<RESOURCE_TYPE, MANAGER_TYPE>(app, fileLocator) \
		{ \
		}

#define END_IMPLEMENT_BASIC_RESOURCE_MANAGER \
	};

// Template fr einfache Ressourcenmanager
template<typename RESOURCE_TYPE, typename DERIVED_MANAGER_TYPE> class BasicResourceManager
	: NonCopyable
	, public ResourceManager
{
public:
	BasicResourceManager(Application& app, FileLocator& fileLocator);

	virtual ~BasicResourceManager() { }

	virtual Resource* getResource(const std::string& filename, const ParameterMap& options = ParameterMap(), bool privateCopy = false);

	virtual void removeResource(Resource* p_resource);

	virtual Handle<RESOURCE_TYPE> getResourceHandle(const std::string& filename, const ParameterMap& options = ParameterMap(), bool privateCopy = false);

protected:
	struct ResourceMapKey
	{
		ResourceMapKey(const std::string& filename,
					   const ParameterMap& options)
			: filename(filename)
			, options(options)
		{
		}

		bool operator <(const ResourceMapKey& rhs) const
		{
			if(filename != rhs.filename) return filename < rhs.filename;
			return options < rhs.options;
		}

		std::string filename;
		ParameterMap options;
	};

	typedef std::map<ResourceMapKey, RESOURCE_TYPE*> ResourceMapType;

	ResourceMapType resources;
};

template<typename RESOURCE_TYPE, typename DERIVED_MANAGER_TYPE> BasicResourceManager<RESOURCE_TYPE, DERIVED_MANAGER_TYPE>::BasicResourceManager(Application& app,
																																				FileLocator& fileLocator)
	: ResourceManager(app, fileLocator)
{

}

template<typename RESOURCE_TYPE, typename DERIVED_MANAGER_TYPE> Resource* BasicResourceManager<RESOURCE_TYPE, DERIVED_MANAGER_TYPE>::getResource(const std::string& filename,
																																				 const ParameterMap& options,
																																				 bool privateCopy)

{
	// Nullzeiger liefern, wenn kein Dateiname angegeben wurde
	if(filename.empty()) return 0;

	// Pfad finden
	const std::string finalPath(fileLocator.findFile(filename));

	// von den Standardoptionen ausgehen und die angegebenen Optionen einfgen
	ParameterMap finalOptions(RESOURCE_TYPE::getDefaultOptions());
	finalOptions.merge(options);

	if(privateCopy)
	{
		// einfach laden, ohne einzutragen
		return new RESOURCE_TYPE(static_cast<DERIVED_MANAGER_TYPE&>(*this), finalPath, finalOptions);
	}
	else
	{
		// nach der Ressource suchen
		ResourceMapKey key(finalPath, options);
		typename ResourceMapType::const_iterator it = resources.find(key);
		if(it != resources.end())
		{
			// Gefunden!
			return it->second;
		}

		// neu laden und eintragen, dann liefern
		RESOURCE_TYPE* p_resource = new RESOURCE_TYPE(static_cast<DERIVED_MANAGER_TYPE&>(*this), finalPath, finalOptions);
		resources[key] = p_resource;
		return p_resource;
	}
}

template<typename RESOURCE_TYPE, typename DERIVED_MANAGER_TYPE> void BasicResourceManager<RESOURCE_TYPE, DERIVED_MANAGER_TYPE>::removeResource(Resource* p_resource)
{
	// nach der Ressource suchen
	for(typename ResourceMapType::iterator it = resources.begin();
		it != resources.end();
		++it)
	{
		if(it->second == p_resource)
		{
			// Und tschss!
			resources.erase(it);
			return;
		}
	}
}

template<typename RESOURCE_TYPE, typename DERIVED_MANAGER_TYPE> Handle<RESOURCE_TYPE> BasicResourceManager<RESOURCE_TYPE, DERIVED_MANAGER_TYPE>::getResourceHandle(const std::string& filename,
																																								   const ParameterMap& options,
																																								   bool unmanagedCopy)
{
	// Referenzzhler manuell erhhen und in ein Handle verpacken
	RESOURCE_TYPE* p_resource = static_cast<RESOURCE_TYPE*>(getResource(filename, options, unmanagedCopy));
	if(p_resource) p_resource->addReference();
	return Handle<RESOURCE_TYPE>(p_resource);
}

#endif
