/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "font.h"

Font::Font(FontManager& manager,
		   const std::string& filename,
		   const ParameterMap& options)
	: Resource(manager, filename, options)
{
	// Font-Definition parsen
	TiXmlDocument doc;
	doc.LoadFile(filename.c_str());

	const TiXmlElement* p_root = doc.FirstChildElement("font");
	const TiXmlElement* p_info = p_root->FirstChildElement("info");
	const TiXmlElement* p_common = p_root->FirstChildElement("common");
	const TiXmlElement* p_pages = p_root->FirstChildElement("pages");
	const TiXmlElement* p_page0 = p_pages->FirstChildElement("page");

	// Name der Textur (Es wird nur eine Seite untersttzt)
	const std::string textureFile(p_page0->Attribute("file"));

	// Textur laden, mit speziellem Filter
	ParameterMap textureOptions;
	textureOptions.set<DownscaleFilter>("downscaleFilter", DF_BOX_ALPHA_WEIGHTED);
	textureHandle = manager.getApplication().getTextureManager().getResourceHandle("fonts/" + textureFile, textureOptions);

	// Gemeinsame Eigenschaften aller Zeichen
	p_common->QueryIntAttribute("base", &base);
	p_common->QueryIntAttribute("scaleH", &scaleH);
	p_common->QueryIntAttribute("scaleW", &scaleW);
	p_common->QueryIntAttribute("lineHeight", &lineHeight);

	// Zeichen-Definition einlesen
	const TiXmlElement* p_chars = p_root->FirstChildElement("chars");
	int id = 0;
	for(const TiXmlElement* p_chr = p_chars->FirstChildElement(); p_chr; p_chr = p_chr->NextSiblingElement()){
		p_chr->QueryIntAttribute("id", &id);
		p_chr->QueryIntAttribute("x", &chars[id].x);
		p_chr->QueryIntAttribute("y", &chars[id].y);
		p_chr->QueryIntAttribute("width", &chars[id].width);
		p_chr->QueryIntAttribute("height", &chars[id].height);
		p_chr->QueryIntAttribute("xoffset", &chars[id].xoffset);
		p_chr->QueryIntAttribute("yoffset", &chars[id].yoffset);
		p_chr->QueryIntAttribute("xadvance", &chars[id].xadvance);
	}
}

const Vec2f Font::drawText(const Vec2f& pos, float size, const RGBA& color, const std::string& text) {

	// Texturierung aktivieren und Textur der Schriftart setzen
	manager.getApplication().setTexturing(true);
	textureHandle->bind();

	// Alpha-Blending aktivieren
	manager.getApplication().setBlending(true);

//	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(pos.x, pos.y, 0);

	glColor4fv(color);

	const float ratio = size / lineHeight;
	unsigned char id = 0;
	Vec2f cursor(0.0f, 0.0f);
	Vec2f textSize(0.0f, 0.0f);
	
	glBegin(GL_QUADS);
	for(size_t i = 0; i < text.length(); ++i) {
		id = text[i];
		if(id == '\n') {
			cursor.x = 0.0f;
			cursor.y += ratio * lineHeight;
		}

		const FontChar& ch = chars[id];

		// links unten
		glTexCoord2i(ch.x, ch.y + ch.height);
		glVertex2f(cursor.x + ratio * ch.xoffset,
			cursor.y + ratio * (ch.yoffset + ch.height));

		// rechts unten
		glTexCoord2i(ch.x + ch.width, ch.y + ch.height);
		glVertex2f(cursor.x + ratio * (ch.xoffset + ch.width),
			cursor.y + ratio * (ch.yoffset + ch.height));

		// rechts oben
		glTexCoord2i(ch.x + ch.width, ch.y);
		glVertex2f(cursor.x + ratio * (ch.xoffset + ch.width),
			cursor.y + ratio * ch.yoffset);

		// links oben
		glTexCoord2i(ch.x, ch.y);
		glVertex2f(cursor.x + ratio * ch.xoffset,
			cursor.y + ratio * ch.yoffset);
		
		cursor.x += ratio * (ch.xadvance - ch.xoffset);
		textSize.x = std::max(textSize.x, cursor.x + ratio * ch.xoffset);
		textSize.y = std::max(textSize.y, cursor.y + ratio * (ch.yoffset + ch.height));
	}
	glEnd();
//	glTranslatef(ratio * (ch.xadvance - ch.xoffset), 0, 0);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

//	glEnable(GL_DEPTH_TEST);

	return textSize;
}

const Vec2f Font::drawCenteredText(const Vec2f& pos,
								   float size,
								   const RGBA& color,
								   const std::string& text)
{
	return drawText(pos - 0.5f * calculateTextSize(size, text), size, color, text);
}

const Vec2f Font::calculateTextSize(float size,
									const std::string& text)
{
	const float ratio = size / lineHeight;
	unsigned char id = 0;
	Vec2f cursor(0.0f, 0.0f);
	Vec2f textSize(0.0f, 0.0f);

	for(size_t i = 0; i < text.length(); ++i) {
		id = text[i];
		if(id == '\n') {
			cursor.x = 0.0f;
			cursor.y += ratio * lineHeight;
		}

		const FontChar& ch = chars[id];
		cursor.x += ratio * (ch.xadvance - ch.xoffset);

		textSize.x = std::max(textSize.x, cursor.x + ratio * ch.xoffset);
		textSize.y = std::max(textSize.y, cursor.y + ratio * (ch.yoffset + ch.height));
	}

	return textSize;
}
