/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "checkpointobject.h"
#include "font.h"
#include "gamestate.h"
#include "highscoremanager.h"
#include "level.h"
#include "playerobject.h"
#include "transitions.h"

void herzentextur(int, int&, int&, int&);

GameState::GameState(Application& app)
	: State("Game", app)
	, p_level(0)
	, paused(false)
	, pointsNeededToWin(0)
	, winner(-1)
	, time(0.0f)
	, startCountDown(0.0f)
	, bestTime(-1.0f)
	, timeToHighscores(0.0f)
	, cameFromHighscores(false)
{
	// HUD-Overlay laden
	hudOverlay = app.getTextureManager().getResourceHandle("sprites/hud.png");

	// Schriftarten laden
	hudFont = app.getFontManager().getResourceHandle("fonts/arial.fnt");
	bigFont = app.getFontManager().getResourceHandle("fonts/bellgothic.fnt");

	// Sounds laden
	winnerSound = app.getSoundManager().getResourceHandle("sounds/winner.ogg");
	loserSound = app.getSoundManager().getResourceHandle("sounds/loser.ogg");
	countDownSound = app.getSoundManager().getResourceHandle("sounds/countdown.ogg");
}

void GameState::onEnter(const ParameterMap& params)
{
	// Spielparameter kopieren
	gameParams = params;

	if(getGameMode() == GM_KOH || getGameMode() == GM_CTG)
	{
		// Anzahl der zum Gewinnen bentigten Punkte holen
		pointsNeededToWin = params.get<int>("pointsNeededToWin");
	}

	// Level laden
	loadLevel(params.get<std::string>("levelFilename"));
}

void GameState::onLeave()
{
	// Level entladen
	loadLevel("");
}

void GameState::onGotFocus()
{
	// Kommen wir aus den Highscores?
	// Wenn ja, dann wird der Level neu gestartet.
	if(cameFromHighscores)
	{
		cameFromHighscores = false;
		loadLevel(levelFilename);
		app.startTransition(new CubeTransition(app, 0.75f));
	}

	// Musik abspielen
	const std::string& musicFilename = p_level->getMusicFilename();
	if(!musicFilename.empty()) app.setMusic(std::string("sounds/") + musicFilename, 0.6f);
}

void GameState::onLostFocus()
{
	// Count-Down-Sound stoppen, falls er noch abgespielt wird
	if(countDownSoundInst && countDownSoundInst->isValid()) countDownSoundInst->stop();
}

void GameState::update()
{
	if(app.wasKeyPressed(SDLK_PAUSE)) paused = !paused;
	const GameMode gameMode = getGameMode();

	if(!paused)
	{
		// Level aktualisieren
		p_level->update();

		// Spielerobjekte holen
		getPlayers();

		// Steuerung der Kamera
		for(int i = 0; i < 2; ++i)
		{
			if(p_player[i])
			{
				// Kamera auf Spieler setzen
				cameraPosition[i] = 0.9f * cameraPosition[i] + 0.1f * p_player[i]->getPosition();

				// Kameradistanz an die Geschwindigkeit anpassen
				const float cameraDistanceTarget = 9.5f + 0.2f * p_player[i]->getBody(0).GetLinearVelocity().Length();
				cameraDistance[i] = 0.95f * cameraDistance[i] + 0.05f * cameraDistanceTarget;
			}
		}

		// Zeit oder Count-Down zhlen
		if(startCountDown > 1.0f)
		{
			if(startCountDown == 4.0f)
			{
				// Sound abspielen
				countDownSoundInst = countDownSound->createInstance();
			}

			startCountDown -= app.getUpdateStepSize();
			if(startCountDown <= 1.0f)
			{
				// Spieler aktivieren, denn es geht los!
				p_player[0]->activate();
				if(p_player[1]) p_player[1]->activate();
			}
		}
		else
		{
			// Die Zeit luft
			time += app.getUpdateStepSize();
			if(startCountDown > 0.0f) startCountDown -= app.getUpdateStepSize();
		}

		// Count-Down zu den Highscores
		if(timeToHighscores > 0.0f)
		{
			timeToHighscores -= app.getUpdateStepSize();
			if(timeToHighscores <= 0.0f)
			{
				HighscoreManager highscoreMgr(app);
				const bool player1InHS = timeNeeded[0] != 0.0f && highscoreMgr.test(levelFilename, ST_TIME, static_cast<int>(timeNeeded[0] * 100.0f));
				const bool player2InHS = timeNeeded[1] != 0.0f && highscoreMgr.test(levelFilename, ST_TIME, static_cast<int>(timeNeeded[1] * 100.0f));

				ParameterMap params;
				params.set<ScoreType>("scoreType", ST_TIME);
				params.set<std::string>("levelFilename", levelFilename);

				if(player1InHS)
				{
					params.set<std::string>("name1", "Spieler 1");
					params.set<unsigned int>("score1", static_cast<unsigned int>(timeNeeded[0] * 100.0f));
				}

				if(player2InHS)
				{
					params.set<std::string>(player1InHS ? "name2" : "name", "Spieler 2");
					params.set<unsigned int>(player1InHS ? "score2" : "score", static_cast<unsigned int>(timeNeeded[1] * 100.0f));
				}

				if(player1InHS || player2InHS)
				{
					app.pushStateByName("Highscore", params);
					app.startTransition(new CubeTransition(app, 0.75f));
					cameFromHighscores = true;
					return;
				}
				else
				{
					// Level neu starten
					loadLevel(levelFilename);
					app.startTransition(new CubeTransition(app, 0.75f));
				}
			}
		}

		if(gameMode == GM_TIME_1P || gameMode == GM_TIME_2P)
		{
			// Hat Spieler 1 das Ziel erreicht?
			const CheckpointObject& finish = *static_cast<const CheckpointObject*>(p_level->getObjectByID("finish"));
			if(!timeNeeded[0] && finish.wasReachedByPlayer(p_player[0]))
			{
				if(winner == -1)
				{
					winner = 0;
					winnerSound->createInstance();
					if(gameMode == GM_TIME_1P) timeToHighscores = 3.0f;
				}
				else
				{
					loserSound->createInstance();
					timeToHighscores = 3.0f;
				}

				timeNeeded[0] = time;
				p_player[0]->deactivate();
			}

			// Hat Spieler 2 das Ziel erreicht?
			if(!timeNeeded[1] && finish.wasReachedByPlayer(p_player[1]))
			{
				if(winner == -1)
				{
					winner = 1;
					winnerSound->createInstance();
				}
				else
				{
					loserSound->createInstance()->fade(1.0f, 0.5f, 4.0f, 1.0f);
					timeToHighscores = 3.0f;
				}

				timeNeeded[1] = time;
				p_player[1]->deactivate();
			}
		}

		if((gameMode == GM_KOH || gameMode == GM_CTG) && winner == -1)
		{
			// Hat ein Spieler durch Punkte gewonnen?
			if(p_player[0]->getPoints() >= pointsNeededToWin) winner = 0;
			else if(p_player[1]->getPoints() >= pointsNeededToWin) winner = 1;

			// Ist ein Spieler (aber nicht beide) tot?
			if(p_player[0]->getHealth() + p_player[1]->getHealth())
			{
				if(!p_player[0]->getHealth()) winner = 1;
				else if(!p_player[1]->getHealth()) winner = 0;
			}

			if(winner != -1)
			{
				// Spielerobjekte deaktivieren
				p_player[0]->deactivate();
				p_player[1]->deactivate();

				// Fanfare abspielen
				winnerSound->createInstance();

				// zu den Highscores (werden bersprungen)
				timeToHighscores = 3.0f;
			}
		}
	}

	if(app.wasKeyPressed(SDLK_RETURN))
	{
		if(gameMode == GM_TIME_2P && winner != -1) timeToHighscores = 0.0001f;
		else if(timeToHighscores > 0.0f) timeToHighscores = 0.0001f;
		else
		{
			// Level neu starten
			loadLevel(levelFilename);
			app.startTransition(new CubeTransition(app, 0.75f));
		}
	}
	else if(app.wasKeyPressed(SDLK_ESCAPE))
	{
		// Zustand verlassen
		paused = false;
		app.popState();
		app.startTransition(new CubeTransition(app, 0.75f));
		return;
	}
}

void GameState::render()
{
	app.clearViewport(RGBA::BLACK);
	const Vec2f screenSize(app.getScreenResolution());

	// Level rendern
	switch(getGameMode())
	{
	case GM_STORY:
	case GM_TIME_1P:
		app.setViewport(screenSize * Vec2f(0.2f, 0.0f), screenSize * Vec2f(0.8f, 1.0f));
		p_level->setCameraPosition(cameraPosition[0]);
		p_level->setCameraDistance(cameraDistance[0]);
		p_level->render();
		break;
	case GM_KOH:
	case GM_CTG:
	case GM_TIME_2P:
		app.setViewport(screenSize * Vec2f(0.2f, 0.0f), screenSize * Vec2f(0.8f, 0.495f));
		p_level->setCameraPosition(cameraPosition[0]);
		p_level->setCameraDistance(cameraDistance[0] - 4.0f);
		p_level->render(p_player[0]);
		app.setViewport(screenSize * Vec2f(0.2f, 0.505f), screenSize * Vec2f(0.8f, 0.495f));
		p_level->setCameraPosition(cameraPosition[1]);
		p_level->setCameraDistance(cameraDistance[1] - 4.0f);
		p_level->render(p_player[1]);
		break;
	}

	// HUD-Overlay rendern
	app.setViewport(Vec2i(0, 0), app.getScreenResolution());
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 1.0, 0.0);
	app.setTexturing(true);
	app.setBlending(true);
	hudOverlay->bind();
	glBegin(GL_QUADS);
	glColor4fv(RGBA::WHITE);
	glTexCoord2i(0, 0);
	glVertex2f(0.0f, 0.0f);
	glTexCoord2i(300, 0);
	glVertex2f(0.22f, 0.0f);
	glTexCoord2i(300, 1024);
	glVertex2f(0.22f, 1.0f);
	glTexCoord2i(0, 1024);
	glVertex2f(0.0f, 1.0f);
	glEnd();

	// HUD rendern
	app.setViewport(Vec2i(0, 0), screenSize * Vec2f(0.2f, 1.0f));
	renderHUD();

	// normalen Viewport wieder setzen
	app.setViewport(Vec2i(0, 0), app.getScreenResolution());
}

GameMode GameState::getGameMode() const
{
	return gameParams.get<GameMode>("gameMode");
}

void GameState::loadLevel(const std::string& filename)
{
	// Count-Down-Sound stoppen, falls er noch abgespielt wird
	if(countDownSoundInst && countDownSoundInst->isValid()) countDownSoundInst->stop();

	if(p_level)
	{
		// alten Level lschen
		delete p_level;
		p_level = 0;
	}

	if(!filename.empty())
	{
		// Level laden
		TiXmlDocument doc;
		doc.LoadFile(app.getFileLocator().findFile(std::string("levels/") + filename).c_str());
		p_level = new Level(app, doc.FirstChildElement("Level"), gameParams);

		// Startzustand
		getPlayers();
		if(p_player[0]) cameraPosition[0] = p_player[0]->getPosition(), cameraDistance[0] = 0.0f;
		if(p_player[1]) cameraPosition[1] = p_player[1]->getPosition(), cameraDistance[1] = 0.0f;
		winner = -1;
		time = 0.0f;
		startCountDown = 0.0f;
		timeNeeded[0] = 0.0f;
		timeNeeded[1] = 0.0f;
		timeToHighscores = 0.0f;
		const GameMode gameMode = getGameMode();

		if(gameMode != GM_STORY)
		{
			// Das Spiel soll erst nach 3 Sekunden losgehen.
			startCountDown = 4.0f;
			p_player[0]->deactivate();
			if(p_player[1])
			{
				p_player[1]->deactivate();

				if(gameMode == GM_TIME_2P)
				{
					// Spieler 2 soll da beginnen, wo auch Spieler 1 beginnt
					p_player[1]->setMirrored(p_player[0]->isMirrored());
					p_player[1]->setPosition(p_player[0]->getPosition());
					p_player[1]->setAngle(p_player[0]->getAngle());
				}
			}

			if(gameMode == GM_TIME_1P || gameMode == GM_TIME_2P)
			{
				// beste Zeit fr den Level abfragen
				HighscoreManager highscoreMgr(app);
				std::vector<Score> highscores(highscoreMgr.get(filename, ST_TIME));
				if(highscores.empty()) bestTime = -1.0f;
				else bestTime = static_cast<float>(highscores[0].value) / 100.0f;
			}
		}
	}

	levelFilename = filename;
}

void GameState::getPlayers()
{
	p_player[0] = p_level->getPlayer(0);
	p_player[1] = p_level->getPlayer(1);
}

void GameState::renderHUD(int life,
						  int points)
{
	static const RGBA textCol(0.15f, 0.65f, 0.82f);
	static const RGBA headingCol(1.0f, 1.0f, 1.0f);
	static const char* p_player1Controls[CM_MAX_VALUE] = { "Pfeiltasten/R.Strg", "Pfeiltasten" };
	static const char* p_player2Controls[CM_MAX_VALUE] = { "W/A/S/D/Leertaste", "W/A/S/D" };
	const GameMode gameMode = getGameMode();
	const ControlMode controlModes[2] = { gameParams.get<ControlMode>("controlMode1"), gameParams.get<ControlMode>("controlMode2") };
	const char* p_playerControls[2] = { p_player1Controls[controlModes[0]], p_player2Controls[controlModes[1]] };

	if(gameMode == GM_STORY)
	{
		app.setTexturing(false);
	//	app.setBlending(false);

		static int herznr = 0;

		static int k = 0;
		static int klife = life;
		++k;
		if(k > 100)
		{
			k = 0;
			klife = MTRand::globalMTRand.randomS32(10, 90);
			herznr = MTRand::globalMTRand.randomS32Exclusive(0, 5);
		}
		
		life = klife;


		float y = 0.848f;
		float breite = 0.3f;

		float hoehe = 0.05f;

		float rahmen = 0.0125f;
		float x = 0 - (breite + rahmen - 0.3f);

		float balken = (life / 100.0f) * breite;

		float grn;
		if(life < 20) 
			grn = 1.0f; 
		else
			grn =0.0f;

		 

		glBegin(GL_QUADS);


		glColor3f ( 0.0f,  0.5f,  1.0f); //hud-bar hintergrund
		glVertex3f(-1.0f,  1.0f,  0.0f);
		glVertex3f( 1.0f,  1.0f,  0.0f);
		glColor3f ( 0.0f,  0.25f,  0.5f);
		glVertex3f( 1.0f,  0.8f,  0.0f);
		glVertex3f(-1.0f,  0.8f,  0.0f);



		glColor4f( 0.0f, 0.0f, 0.0f, 1.0f);  // rahmen
		glVertex3f(x + breite + rahmen, y + hoehe + rahmen, 0.0f);
		glVertex3f(x + breite + rahmen, y - rahmen, 0.0f);
		glVertex3f(x - rahmen, y - rahmen, 0.0f);
		glVertex3f(x - rahmen, y + hoehe + rahmen, 0.0f);


		glColor4f(1.0f, 1.0f, 1.0f, 1.0f); //hintergrund
		glVertex3f(x + breite, y + hoehe, 0.0f);
		glColor4f(0.8f, 0.8f, 0.8f, 1.0f);
		glVertex3f(x + breite, y, 0.0f);
		glVertex3f(x, y , 0.0f);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glVertex3f(x , y + hoehe, 0.0f);

		glColor4f(1.0f * (100.0f/life), 1.0f * (life/100.0f), 0.0f, 1.0f);  //balken
		glVertex3f(x + breite, y + hoehe, 0.0f);
		glVertex3f(x + breite, y, 0.0f);
		glColor4f(grn, 1.0f - grn, 0.0f, 1.0f);
		glVertex3f(x + (breite - balken), y, 0.0f);
		glVertex3f(x + (breite - balken), y + hoehe, 0.0f);



		glColor4f ( 1.0f,  1.0f,  1.0f, 0.55f); //highlight
		glVertex3f(-0.995f,  0.992f,  0.0f);
		glVertex3f( 0.995f,  0.992f,  0.0f);
		glColor4f ( 1.0f,  1.0f,  1.0f, 0.15f);
		glVertex3f( 0.995f,  0.88f,  0.0f);
		glVertex3f(-0.995f,  0.88f,  0.0f);

		glEnd();

		static TextureHandle herz = app.getTextureManager().getResourceHandle("sprites/heart.png");
		//p_level->addResourceToCache(herz);

		herz->bind();

		app.setTexturing(true);
		app.setBlending(true);

		glColor4fv(RGBA::WHITE);

	//gluPerspective(90.0, app.getScreenAspect(), 0.001, 1000.0);


		glTranslatef(0.2f, 0.95f, 0.0f);

		glBegin(GL_QUADS);

	//	glColor4f(0.0, 0.0, 0.0, 1.0);
		static int herzlinks, herzmitte, herzrechts;

		herzentextur(life, herzlinks, herzmitte, herzrechts);
		
		//herz mitte ganz
		glTexCoord2i(herzmitte, 0); // links oben
		glVertex2f(-0.03f, 0.03f * app.getViewportAspect());

		glTexCoord2i(herzmitte + 32, 0); // rechts oben
		glVertex2f(0.03f, 0.03f * app.getViewportAspect());
		
		glTexCoord2i(herzmitte + 32, 32); // rechts unten
		glVertex2f(0.03f, -0.03f * app.getViewportAspect());

		glTexCoord2i(herzmitte, 32); // links unten
		glVertex2f(-0.03f, -0.03f * app.getViewportAspect());

		//herz links ganz
		glTexCoord2i(herzlinks, 0); // links oben
		glVertex2f(-0.03f - 0.06f, 0.03f * app.getViewportAspect());

		glTexCoord2i(herzlinks + 32, 0); // rechts oben
		glVertex2f(0.03f - 0.06f, 0.03f * app.getViewportAspect());

		glTexCoord2i(herzlinks + 32, 32); // rechts unten
		glVertex2f(0.03f - 0.06f, -0.03f * app.getViewportAspect());

		glTexCoord2i(herzlinks, 32); // links unten
		glVertex2f(-0.03f - 0.06f, -0.03f * app.getViewportAspect());

		//herz rechts teil
		glTexCoord2i(herzrechts, 0); // links oben
		glVertex2f(-0.03f + 0.06f, 0.03f * app.getViewportAspect());

		glTexCoord2i(herzrechts + 32, 0); // rechts oben
		glVertex2f(0.03f + 0.06f, 0.03f * app.getViewportAspect());

		glTexCoord2i(herzrechts + 32, 32); // rechts unten
		glVertex2f(0.03f + 0.06f, -0.03f * app.getViewportAspect());

		glTexCoord2i(herzrechts, 32); // links unten
		glVertex2f(-0.03f + 0.06f, -0.03f * app.getViewportAspect());



		glEnd();

		glTranslatef(-0.2f, -0.95f, 0.0f);

		// Projektionsmatrix einstellen
		glMatrixMode(GL_PROJECTION);
		glOrtho(0.0, 32.0, 24.0, 0.0, 0.0, 1.0);
		
		glMatrixMode(GL_MODELVIEW);

		hudFont->drawText(Vec2f(0.0f, 0.0f), 1.0f, RGBA(1.0f, 1.0f, 0.0f), "Time:");
		hudFont->drawText(Vec2f(14.5f, 0.0f), 1.0f, RGBA(1.0f, 1.0f, 0.0f), "Life:");
		hudFont->drawText(Vec2f(28.8f, 0.0f), 1.0f, RGBA(1.0f, 1.0f, 0.0f), "Points:");

		int timeRemaining = 99 - int(time);
		
		hudFont->drawText(Vec2f(0.0f, 1.0f), 1.0f, RGBA(1.0f, 1.0f, 0.0f), toString(timeRemaining));
		hudFont->drawText(Vec2f(14.5f, 1.0f), 1.0f, RGBA(1.0f, 1.0f, 0.0f), toString(life));
		
		
		PlayerObject* p_player = dynamic_cast<PlayerObject*>(p_level->getObjectByID("player1"));
		hudFont->drawText(Vec2f(28.8f, 1.0f), 1.0f, RGBA(1.0f, 1.0f, 0.0f), toString(p_player->getPoints()));

	//	hudFont->drawText(Vec2f(29.0f, 3.0f), 1.0f, RGBA(0.0f, 0.5f, 1.0f), toString(balken));



		// x = 31 ist das letzte dargestellte zeichen
		// y = 23 ist die letzte spalte
	}
	else if(gameMode == GM_KOH || gameMode == GM_CTG)
	{
		if(startCountDown > 0.0f) drawStartCountDown();

		// Projektion einstellen
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluOrtho2D(0.0, app.getViewportAspect(), 1.0, 0.0);

		float yPos = 0.13f;
		hudFont->drawText(Vec2f(0.02f, yPos), 0.05f, headingCol, "Spieler 1"); yPos += 0.055f;
		hudFont->drawText(Vec2f(0.02f, yPos), 0.03f, textCol, p_playerControls[0]); yPos += 0.035f;
		Vec2f textSize(bigFont->drawText(Vec2f(0.02f, yPos), 0.1f, textCol, toString(p_player[0]->getPoints())));
		hudFont->drawText(Vec2f(0.02f + textSize.x, yPos + textSize.y - 0.035f), 0.035f, textCol, std::string("/") + toString(pointsNeededToWin));

/*		std::ostringstream s;
		s << std::endl << "Pfeiltasten" << std::endl << "Punkte: " << p_player[0]->getPoints() << " (" << pointsNeededToWin << ")";
		s << std::endl << "Status: " << p_player[0]->getHealth() << '%';
		hudFont->drawText(Vec2f(0.02f, 0.06f), 0.035f, textCol, s.str());*/

		yPos = 0.68f;
		hudFont->drawText(Vec2f(0.02f, yPos), 0.05f, headingCol, "Spieler 2"); yPos += 0.055f;
		hudFont->drawText(Vec2f(0.02f, yPos), 0.03f, textCol, p_playerControls[1]); yPos += 0.035f;
		textSize = bigFont->drawText(Vec2f(0.02f, yPos), 0.1f, textCol, toString(p_player[1]->getPoints()));
		hudFont->drawText(Vec2f(0.02f + textSize.x, yPos + textSize.y - 0.035f), 0.035f, textCol, std::string("/") + toString(pointsNeededToWin));

/*		s.str("");
		s << std::endl << "W, A, S, D" << std::endl << "Punkte: " << p_player[1]->getPoints() << " (" << pointsNeededToWin << ")";
//		s << std::endl << "Status: " << p_player[1]->getHealth() << '%';
		hudFont->drawText(Vec2f(0.02f, 0.56f), 0.035f, textCol, s.str());*/

		if(winner != -1)
		{
			// Projektion neu einstellen
			app.setViewport(Vec2i(0, 0), app.getScreenResolution());
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			gluOrtho2D(0.0, app.getViewportAspect(), 1.0, 0.0);
			const float centerX = app.getViewportAspect() * 0.6f;

			// Schriftzug rendern
			bigFont->drawCenteredText(Vec2f(centerX, 0.5f), 0.15f,
									  RGBA(0.0f, 1.0f, 0.0f, 0.75f + 0.25f * sinf(10.0f * app.getTime())),
									  winner == 0 ? "SPIELER 1\nGEWINNT!" : "SPIELER 2\nGEWINNT!");
		}
	}
	else if(gameMode == GM_TIME_1P || gameMode == GM_TIME_2P)
	{
		// Projektion einstellen
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluOrtho2D(0.0, app.getViewportAspect(), 1.0, 0.0);

		float yPos = 0.13f;
		hudFont->drawText(Vec2f(0.02f, yPos), 0.05f, headingCol, "Spieler 1"); yPos += 0.055f;
		hudFont->drawText(Vec2f(0.02f, yPos), 0.03f, textCol, p_playerControls[0]); yPos += 0.035f;
		bigFont->drawText(Vec2f(0.02f, yPos), 0.07f, timeNeeded[0] ? RGBA::RED : textCol, formatTime(timeNeeded[0] ? timeNeeded[0] : time));

		if(bestTime != -1.0f)
		{
			// beste Zeit anzeigen
			yPos = 0.4575f;
			hudFont->drawText(Vec2f(0.02f, yPos), 0.035f, headingCol, "Rekordzeit"); yPos += 0.035f;
			bigFont->drawText(Vec2f(0.02f, yPos), 0.05f, textCol, formatTime(bestTime));
		}

		if(gameMode == GM_TIME_2P)
		{
			yPos = 0.68f;
			hudFont->drawText(Vec2f(0.02f, yPos), 0.05f, headingCol, "Spieler 2"); yPos += 0.055f;
			hudFont->drawText(Vec2f(0.02f, yPos), 0.03f, textCol, p_playerControls[1]); yPos += 0.035f;
			bigFont->drawText(Vec2f(0.02f, yPos), 0.07f, timeNeeded[1] ? RGBA::RED : textCol, formatTime(timeNeeded[1] ? timeNeeded[1] : time));
		}

		if(startCountDown > 0.0f) drawStartCountDown();
		else if(gameMode == GM_TIME_2P && winner != -1)
		{
			// Projektion neu einstellen
			app.setViewport(Vec2i(0, 0), app.getScreenResolution());
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			gluOrtho2D(0.0, app.getViewportAspect(), 1.0, 0.0);
			const float centerX = app.getViewportAspect() * 0.6f;

			// Schriftzug rendern
			bigFont->drawCenteredText(Vec2f(centerX, 0.5f), 0.15f,
									  RGBA(0.0f, 1.0f, 0.0f, 0.75f + 0.25f * sinf(10.0f * app.getTime())),
									  winner == 0 ? "SPIELER 1\nGEWINNT!" : "SPIELER 2\nGEWINNT!");
		}
	}

	if(winner == -1 && !p_player[0]->getHealth() && (!p_player[1] || !p_player[1]->getHealth())) {
		// beide Spieler tot

		// Projektion neu einstellen
		app.setViewport(Vec2i(0, 0), app.getScreenResolution());
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluOrtho2D(0.0, app.getViewportAspect(), 1.0, 0.0);
		const float centerX = app.getViewportAspect() * 0.6f;

		// Schriftzug rendern
		bigFont->drawCenteredText(Vec2f(centerX, 0.5f), 0.15f,
								  RGBA(1.0f, 0.0f, 0.0f, 0.75f + 0.25f * sinf(10.0f * app.getTime())),
								  "GAME OVER");
	}

}

void GameState::drawStartCountDown()
{
	// Projektion neu einstellen
	app.setViewport(Vec2i(0, 0), app.getScreenResolution());
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, app.getViewportAspect(), 1.0, 0.0);
	const float centerX = app.getViewportAspect() * 0.6f;

	// Schriftzug rendern
	const float size = 1.0f - startCountDown + floorf(startCountDown);
	RGBA color;
	if(startCountDown < 1.0f) color = RGBA(0.0f, 1.0f, 0.0f, 1.0f - size);
	else if(startCountDown < 2.0f) color = RGBA(1.0f, 1.0f, 0.0f, 1.0f - size);
	else if(startCountDown < 3.0f) color = RGBA(1.0f, 0.5f, 0.0f, 1.0f - size);
	else color = RGBA(1.0f, 0.0f, 0.0f, 1.0f - size);
	bigFont->drawCenteredText(Vec2f(centerX, 0.5f), size, color,
							  startCountDown < 1.0f ? "LOS!" : toString(static_cast<int>(startCountDown)));
}

void herzentextur(int life, int &links, int &mitte, int &rechts)
{
	enum stand
	{
		voll = 0,
		dreiviertel = 32,
		halb = 64,
		einviertel = 96,
		leer = 128
	};

	if(life < 41)
	{
		mitte = rechts = leer;
		if(life > 33)
		{
			links = voll;
		}
		else if(life > 25)
		{
			links = dreiviertel;
		} 
		else if(life > 16)
		{
			links = halb;
		}
		else if(life > 8)
		{
			links = einviertel;
		}
		else
		{
			links = leer;
		}
	}
	else if(life < 67)
	{
		links = voll;
		rechts = leer;
		if(life > 58)
		{
			mitte = voll;
		}
		else if(life > 50)
		{
			mitte = dreiviertel;
		}
		else if(life > 41)
		{
			mitte = halb;
		}
		else if(life > 33)
		{
			mitte = einviertel;
		}
		else
		{
			mitte = leer;
		}
	}
	else
	{
		links = mitte = voll;
		if(life > 91)
		{
			rechts = voll;
		}
		else if(life > 83)
		{
			rechts = dreiviertel;
		}
		else if(life > 75)
		{
			rechts = halb;
		}
		else if(life > 66)
		{
			rechts = einviertel; 
		}
		else
		{
			rechts = leer;
		}
	}
}
