/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "guhaselnutobject.h"
#include "level.h"
#include "playerobject.h"

GuhaselnutObject::GuhaselnutObject(Layer& layer,
								   const TiXmlElement* p_xml,
								   const ParameterMap& gameParams)
	: Object(layer, p_xml, gameParams)
	, respawnTime(5.0f)
	, p_carrier(0)
	, spawnCountDown(0.0f)
{
	visibilityRadius = 1.5f;

	// Dieses Objekt soll nur im "Capture the Guhaselnut"-Modus existieren.
	if(getGameMode() != GM_CTG) kill();

	// Respawn-Zeit holen
	p_xml->QueryFloatAttribute("respawnTime", &respawnTime);

	// Textur und Sounds holen
	level.addResourceToCache(sprites = app.getTextureManager().getResourceHandle("sprites/guhaselnut.png"));
	level.addResourceToCache(takeSound = app.getSoundManager().getResourceHandle("sounds/nut-take.ogg"));
	level.addResourceToCache(captureSound = app.getSoundManager().getResourceHandle("sounds/nut-capture.ogg"));
	level.addResourceToCache(spawnSound = app.getSoundManager().getResourceHandle("sounds/nut-spawn.ogg"));

	b2BodyDef bodyDef;
	b2Body& body = createBody(bodyDef);
	b2CircleDef circleDef;
	circleDef.radius = 0.5f;
	circleDef.isSensor = true;
	body.CreateShape(&circleDef);
	body.SetMassFromShapes();

	// Initialisierung fertigstellen
	finishInit(p_xml);

	// Position merken
	spawnPosition = getPosition();
}

void GuhaselnutObject::update()
{
	// die Nuss folgt dem Spieler, der sie trgt
	if(p_carrier)
	{
		const Vec2f currentPosition(getPosition());
		const Vec2f targetPosition(p_carrier->getPosition() + Vec2f(0.0f, -2.5f));
		setPosition(currentPosition * 0.8f + targetPosition * 0.2f);
		setAngle(getAngle() + 4.0f * app.getUpdateStepSize());
	}

	if(spawnCountDown > 0.0f)
	{
		// Count-Down
		spawnCountDown -= app.getUpdateStepSize();
		if(spawnCountDown <= 0.0f)
		{
			// wieder auftauchen
			setPosition(spawnPosition);
			setAngle(0.0f);

			// Sound abspielen
			spawnSound->createInstance();
		}
	}
}

void GuhaselnutObject::render()
{
	if(spawnCountDown > 0.0f) return;

	app.setBlending(true);
	app.setTexturing(true);
	sprites->bind();
	drawBodyQuad(0, Vec2i(0, 0), Vec2i(128, 128), Vec2f(2.0f, 2.0f));
}

void GuhaselnutObject::onCollision(const Contact& contact)
{
	if(spawnCountDown > 0.0f) return;

	if(PlayerObject* p_player = dynamic_cast<PlayerObject*>(contact.p_otherObject))
	{
		if(p_carrier != p_player)
		{
			// Trger ndern und Sound abspielen
			p_carrier = p_player;
			takeSound->createInstance(true, false, 1.0f, MTRand::globalMTRand.randomFloat(0.9f, 1.1f));
		}
	}
}

void GuhaselnutObject::onCapture(PlayerObject& player)
{
	if(spawnCountDown > 0.0f) return;

	// Der Spieler bekommt Punkte!
	player.addPoints(1);

	// verschwinden und erst nach einiger Zeit wieder auftauchen
	spawnCountDown = respawnTime;

	// Niemand hat die Nuss!
	p_carrier = 0;

	// Sound abspielen
	captureSound->createInstance();
}

PlayerObject& GuhaselnutObject::getCarrier() const
{
	return *p_carrier;
}
