/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mail‰nder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "highscoremanager.h"

HighscoreManager::HighscoreManager(Application& app)
	: app(app)
	, maxEntries(10)
{
	load();
}

void HighscoreManager::add(std::string level, ScoreType scoreType, std::string name, HighscoreValueType value)
{
	if(!test(level, scoreType, value)) return;
	Score score;
	score.name = name;
	score.value = value;
	scores[level][scoreType].push_back(score);

	// sortieren
	if(scoreType == ST_POINTS) std::sort(scores[level][scoreType].begin(), scores[level][scoreType].end(), ScoreSorterDesc());
	else if(scoreType == ST_TIME) std::sort(scores[level][scoreType].begin(), scores[level][scoreType].end(), ScoreSorterAsc());

	// Einträge, die zu viel sind, fliegen raus
	while(scores[level][scoreType].size() > maxEntries) scores[level][scoreType].pop_back();
}

std::vector<Score> HighscoreManager::get(std::string level, ScoreType scoreType)
{
	if(scoreType == ST_NULL) return std::vector<Score>();
	else return scores[level][scoreType];
}

void HighscoreManager::load()
{
	// Highscore-Datei einlesen
	TiXmlDocument document;
	if(!document.LoadFile(app.getFileLocator().findFile("highscore.xml").c_str())) return;
	
	TiXmlElement* p_highscore = document.FirstChildElement("Highscore");
	if(!p_highscore) return;
	
	// alle Level durchgehen
	for(const TiXmlElement* p_level = p_highscore->FirstChildElement("Level");
		p_level;
		p_level = p_level->NextSiblingElement("Level"))
	{
		std::string filename(p_level->Attribute("filename"));
		// alle ScoreTypes durchgehen
		for(const TiXmlElement* p_scoretype = p_level->FirstChildElement("ScoreType");
			p_scoretype;
			p_scoretype = p_scoretype->NextSiblingElement("ScoreType"))
		{
			ScoreType scoretype = ScoreType(fromString<int>(p_scoretype->Attribute("type")));
			// alle Punkte einlesen
			for(const TiXmlElement* p_score = p_scoretype->FirstChildElement("Score");
				p_score;
				p_score = p_score->NextSiblingElement("Score"))
			{
				Score score;
				score.name = p_score->Attribute("name");
				score.value = fromString<HighscoreValueType>(p_score->Attribute("value"));
				scores[filename][scoretype].push_back(score);
			}
		}
	}
}

void HighscoreManager::save()
{
	TiXmlDocument document;
	
	// XML-Deklaration
	TiXmlDeclaration* decl = new TiXmlDeclaration("1.0", "", "");
	document.LinkEndChild(decl);
	
	// Root-Element
	TiXmlElement* p_root = new TiXmlElement("Highscore");

	for(ScoreTableType::const_iterator levelIT = scores.begin();
		levelIT != scores.end();
		++levelIT)
	{
		// Level
		TiXmlElement* p_level = new TiXmlElement("Level");
		p_level->SetAttribute("filename", levelIT->first.c_str());
		
		for(STListType::const_iterator scoretypeIT = levelIT->second.begin();
			scoretypeIT != levelIT->second.end();
			++scoretypeIT)
		{
			// ScoreType
			TiXmlElement* p_scoretype = new TiXmlElement("ScoreType");
			p_scoretype->SetAttribute("type", toString<int>(int(scoretypeIT->first)).c_str());
			
			for(ScoreListType::size_type i = 0; i < scoretypeIT->second.size(); ++i) {
				// Score
				TiXmlElement* p_score = new TiXmlElement("Score");
				p_score->SetAttribute("name", scoretypeIT->second[i].name.c_str());
				p_score->SetAttribute("value", toString<HighscoreValueType>(scoretypeIT->second[i].value).c_str());
				
				p_scoretype->LinkEndChild(p_score);
			}
			
			p_level->LinkEndChild(p_scoretype);
		}
		
		p_root->LinkEndChild(p_level);
	}

	document.LinkEndChild(p_root);
	
	// @TODO: Prüfsumme
	document.SaveFile(app.getFileLocator().findFile("highscore.xml").c_str());
}

bool HighscoreManager::test(std::string level, ScoreType scoreType, HighscoreValueType value)
{
	ScoreListType scoreList = get(level, scoreType);
	if(!value || scoreList.size() < maxEntries) return true;
	if(scoreType == ST_POINTS && value > scoreList.back().value) return true;
	if(scoreType == ST_TIME && value < scoreList.back().value) return true;
	return false;
}

int HighscoreManager::getMaxEntries()
{
	return static_cast<int>(maxEntries);
}
