/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mail‰nder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "highscorestate.h"
#include "transitions.h"
#include "utils.h"

HighscoreState::HighscoreState(Application& app)
	: State("Highscore", app)
	, submittedScore(0)
	, submittedScore2(0)
	, manager(app)
	, cursorBlinkSpeed(M_PI_FLOAT)
	, maxNameLength(strlen("Jan Hendrik Hapke"))
	
{
	// Schriftarten laden
	font = app.getFontManager().getResourceHandle("fonts/arial.fnt");
	bigFont = app.getFontManager().getResourceHandle("fonts/bellgothic.fnt");
}

void HighscoreState::onEnter(const ParameterMap& params)
{
	levelFilename = params.get<std::string>("levelFilename");
	scoreType = params.get<ScoreType>("scoreType");

	submittedScore = 0;
	submittedScore2 = 0;
	input = "";
	input2 = "";
	playerEntering = 0;

	// sollen Punkte eingetragen werden?
	params.query<unsigned int>("score1", submittedScore);
	params.query<unsigned int>("score2", submittedScore2);
	// Ist der Name bereits bekannt?
	if(lastUsedName[0].empty()) params.query<std::string>("name1", input); else input = lastUsedName[0];
	if(lastUsedName[1].empty()) params.query<std::string>("name2", input2); else input2 = lastUsedName[1];
	
	if(!manager.test(levelFilename, scoreType, submittedScore))
	{
		// Punktzahl reicht nicht für Highscore
		submittedScore = 0;
		input = "";
	}

	if(!manager.test(levelFilename, scoreType, submittedScore2))
	{
		// Punktzahl reicht nicht für Highscore
		submittedScore2 = 0;
		input2 = "";
	}
	
	scores = manager.get(levelFilename, scoreType);
}

void HighscoreState::onLeave()
{
	input = "";
	submittedScore = 0;
}

void HighscoreState::onGotFocus()
{
}

void HighscoreState::onLostFocus()
{
}

void HighscoreState::update()
{
	if(scoreType == ST_NULL || app.wasKeyPressed(SDLK_ESCAPE) ||
	   (!submittedScore && app.wasKeyPressed(SDLK_RETURN)))
	{
		app.popState();
		app.startTransition(new CubeTransition(app, 0.75f));
		return;
	}

	if(!submittedScore) return;
	
	// Eingabe verarbeiten
	SDL_keysym key;
	while(app.hasMoreKeys())
	{
		key = app.popKey();
		switch (key.sym)
		{
		case SDLK_BACKSPACE:
			input = input.substr(0, input.length() - 1);
			break;
		case SDLK_RETURN:
			if(strip(input) != "")
			{
				manager.add(levelFilename, scoreType, input, submittedScore);
				manager.save();
				scores = manager.get(levelFilename, scoreType);
				lastUsedName[playerEntering] = input;

				if(submittedScore2)
				{
					input = input2;
					submittedScore = submittedScore2;
					input2 = "";
					submittedScore2 = 0;
					++playerEntering;
				}
				else
				{
					input = "";
					submittedScore = 0;
				}
			}
			break;
		default:
			if(input.length() < maxNameLength) input += char(key.unicode);
			break;
		}
	}

}

void HighscoreState::render()
{
	static const RGBA textCol(0.15f, 0.65f, 0.82f);
	static const RGBA headingCol(1.0f, 1.0f, 1.0f);
	static const RGBA newEntryCol(1.0f, 0.5f, 0.1f);

	if(scoreType == ST_NULL)
	{
		app.popState();
		app.startTransition(new CubeTransition(app, 0.75f));
		return;
	}

	app.clearViewport(RGBA::BLACK);

	// Projektionsmatrix einstellen
	glMatrixMode(GL_PROJECTION);
	glOrtho(0.0, 32.0, 24.0, 0.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	
	bigFont->drawCenteredText(Vec2f(16.0f, 2.0f), 2.0f, headingCol, "Highscore");
	bigFont->drawCenteredText(Vec2f(16.0f, 22.0f), 1.5f, headingCol, "Eingabe/Weiter: Return");

	float yPos = 4.0f;
	float xPos = 8.0f;
	
	bool inputDrawn = false;
	float inputYPos = yPos;
	float inputXPos = xPos;
	
	int inputRank = 0;
	int rank = 1;
	
	for(ScoreListType::size_type i = 0; i < scores.size() && rank <= manager.getMaxEntries(); ++i)
	{
		Score score = scores[i];
		if(submittedScore && !inputDrawn &&
			((scoreType == ST_POINTS && submittedScore > score.value) ||
			 (scoreType == ST_TIME && submittedScore < score.value)))
		{
			// der neue Punktestand muss über diesem eingetragen werden
			yPos += 1.5f;
			inputDrawn  = true;
			inputRank = rank;
			++rank;
		}

		if(rank > manager.getMaxEntries()) break;

		std::ostringstream ss;
		ss << (rank) << ". " << score.name;
		font->drawText(Vec2f(xPos, yPos), 1.0f, rank == manager.getMaxEntries() + 1 ? RGBA(0.5f) : textCol, ss.str());

		std::string theScore;
		switch(scoreType)
		{
		case ST_POINTS:
			theScore = toString(score.value);
			break;
		case ST_TIME:
			theScore = formatTime(static_cast<float>(score.value) / 100.0f);
			break;
		}

		font->drawText(Vec2f(xPos + 10.0f, yPos), 1.0f, textCol, theScore);
		yPos += 1.5f;
		
		if(!inputDrawn)
		{
			// Position für neuen Punktestand aktualisieren
			inputYPos = yPos;
			inputXPos = xPos;
		}
		++rank;
	}
	
	if(submittedScore) {
	
		if(!inputRank) inputRank = rank;
		std::ostringstream ss;
		ss << inputRank << ". " << input;
		
		// bisher eingegebenen Text und Cursor rendern
		if(static_cast<int>(app.getTime() * cursorBlinkSpeed) & 1) ss << '_';
		font->drawText(Vec2f(inputXPos, inputYPos), 1.0f, newEntryCol, ss.str());

		std::string theScore;
		switch(scoreType)
		{
		case ST_POINTS:
			theScore = toString(submittedScore);
			break;
		case ST_TIME:
			theScore = formatTime(static_cast<float>(submittedScore) / 100.0f);
			break;
		}

		// Punktezahl
		font->drawText(Vec2f(inputXPos + 10.0f, inputYPos), 1.0f, newEntryCol, theScore);
		yPos += 1.5f;
	}

}

