/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "kingareaobject.h"
#include "level.h"
#include "playerobject.h"

KingAreaObject::KingAreaObject(Layer& layer,
							   const TiXmlElement* p_xml,
							   const ParameterMap& gameParams)
	: Object(layer, p_xml, gameParams)
	, size(0.0f, 0.0f)
	, color(1.0f, 0.0f, 0.0f, 0.5f)
{
	// Dieses Objekt soll nur im "King of the Hill"-Modus existieren.
	if(getGameMode() != GM_KOH) kill();

	// Textur holen
	level.addResourceToCache(sprites = app.getTextureManager().getResourceHandle("sprites/area.png"));

	// Gre lesen
	if(const char* p_size = p_xml->Attribute("size")) size = fromString<Vec2f>(p_size);
	visibilityRadius = std::max(size.x, size.y);

	// Physik
	b2BodyDef bodyDef;
	b2Body& body = createBody(bodyDef);
	b2PolygonDef polygonDef;
	polygonDef.SetAsBox(size.x * 0.5f, size.y * 0.5f);
	polygonDef.isSensor = true;
	body.CreateShape(&polygonDef);
	createBody(bodyDef);

	// Initialisierung fertigstellen
	finishInit(p_xml);
}

void KingAreaObject::update()
{
	RGBA targetColor = playersInside.empty() ? RGBA(1.0f, 0.0f, 0.0f, 0.5f) : RGBA(1.0f, 1.0f, 0.0f, 0.5f);

	// ist mehr als ein Spieler im Bereich, dann werden alle Zhler zurckgesetzt
	if(playersInside.size() > 1) playerTimes.clear();

	// neue Spieler zur Map hinzufgen
	for(PlayerSetType::const_iterator it = playersInside.begin();
		it != playersInside.end();
		++it)
	{
		if(playerTimes.find(*it) == playerTimes.end())
		{
			// mit einer Zeit von 0 starten
			playerTimes[*it] = 0.0f;
		}
	}

	// Zeit der Spieler erhhen, die im Bereich sind,
	// und die Zeit der Spieler, die nicht im Bereich sind, zurcksetzen.
	// Ist ein Spieler lange genug im Bereich und kein anderer,
	// dann bekommt er Punkte.
	for(PlayerTimeMapType::iterator it = playerTimes.begin();
		it != playerTimes.end();
		++it)
	{
		if(playersInside.find(it->first) != playersInside.end())
		{
			// Spieler ist drin!
			it->second += app.getUpdateStepSize();

			// Punkte?
			if(it->second >= 1.0f && playersInside.size() == 1)
			{
				it->first->addPoints(app.getUpdateStepSize());
				targetColor = RGBA(0.0f, 1.0f, 0.0f, 0.5f);
			}
		}
		else
		{
			// Spieler ist nicht drin!
			it->second = 0.0f;
		}
	}

	// Farbe der Zielfarbe anpassen
	color = 0.99f * color + 0.01f * targetColor;

	// Liste der Spieler, die im Bereich sind, zurcksetzen
	playersInside.clear();
}

void KingAreaObject::render()
{
	app.setBlending(true);
	app.setTexturing(true);
	sprites->bind();

	RGBA theColor(color);
	theColor.a *= 0.75f + 0.25f * sinf(20.0f * app.getTime());
	drawBodyQuad(0, Vec2i(0, 0), Vec2i(128, 128), size, false, false, theColor);
}

void KingAreaObject::onCollision(const Contact& contact)
{
	if(PlayerObject* p_player = dynamic_cast<PlayerObject*>(contact.p_otherObject))
	{
		// Spieler hinzufgen
		playersInside.insert(p_player);
	}
}
