/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "checkpointobject.h"
#include "dummyobject.h"
#include "guhaselnutobject.h"
#include "homebaseobject.h"
#include "kingareaobject.h"
#include "layer.h"
#include "level.h"
#include "playerobject.h"
#include "simpleobject.h"
#include "textobject.h"

Layer::Layer(Level& level,
			 int index,
			 const TiXmlElement* p_xml,
			 const ParameterMap& gameParams)
	: level(level)
	, app(level.getApplication())
	, index(index)
	, z(1.0f)
	, color(RGBA::WHITE)
	, blendMode(1)
	, visible(true)
	, updatingObjects(false)
{
	// ID lesen oder automatisch generieren
	if(const char* p_id = p_xml->Attribute("id")) id = p_id;
	else id = autoIDString<Layer>("Layer#");

	// z-Wert abfragen
	p_xml->QueryFloatAttribute("z", &z);

	// Farbe abfragen
	if(const char* p_color = p_xml->Attribute("color")) color = fromString<RGBA>(p_color);

	// Blending-Modus lesen
	p_xml->QueryIntAttribute("blendMode", &blendMode);

	// Sichtbarkeit lesen
	int visible = 1;
	p_xml->QueryIntAttribute("visible", &visible);
	this->visible = visible != 0;

	// Objekte lesen
	const TiXmlElement* p_objectsXML = p_xml->FirstChildElement("Objects");
	if(p_objectsXML)
	{
		for(const TiXmlElement* p_objectXML = p_objectsXML->FirstChildElement();
			p_objectXML;
			p_objectXML = p_objectXML->NextSiblingElement())
		{
			Object* p_object = 0;

			// Typ des Objekts?
			const std::string objectType(p_objectXML->Value());
			if(objectType == "CheckpointObject") p_object = new CheckpointObject(*this, p_objectXML, gameParams);
			else if(objectType == "DummyObject") p_object = new DummyObject(*this, p_objectXML, gameParams);
			else if(objectType == "GuhaselnutObject") p_object = new GuhaselnutObject(*this, p_objectXML, gameParams);
			else if(objectType == "HomebaseObject") p_object = new HomebaseObject(*this, p_objectXML, gameParams);
			else if(objectType == "KingAreaObject") p_object = new KingAreaObject(*this, p_objectXML, gameParams);
			else if(objectType == "PlayerObject") p_object = new PlayerObject(*this, p_objectXML, gameParams);
			else if(objectType == "SimpleObject") p_object = new SimpleObject(*this, p_objectXML, gameParams);
			else if(objectType == "TextObject") p_object = new TextObject(*this, p_objectXML, gameParams);
			else
			{
				// ungltiger Typ
				assert(false);
			}

			// Objekt hinzufgen, falls es noch lebt, sonst wieder lschen
			if(p_object->isAlive()) addObject(*p_object);
			else delete p_object;
		}
	}
}

Layer::~Layer()
{
	// alle Objekte lschen
	for(ObjectMapType::const_iterator it = objects.begin();
		it != objects.end();
		++it)
	{
		delete it->second;
	}
}

void Layer::addObject(Object& object)
{
	if(updatingObjects)
	{
		ObjectListEntryType entry;
		entry.p_obj = &object;
		entry.alsoDelete = true;
		// spter hinzufgen
		objectsToAdd.push_back(entry);
	}
	else
	{
		objects[object.getID()] = &object;
		object.onAdd();
	}
}

void Layer::removeObject(Object& object)
{
	removeObject(object, true);
}

void Layer::removeObject(Object& object, bool alsoDelete)
{
	if(updatingObjects)
	{
		ObjectListEntryType entry;
		entry.alsoDelete = alsoDelete;
		entry.p_obj = &object;
		// spter entfernen
		objectsToRemove.push_back(entry);
	}
	else
	{
		objects.erase(object.getID());
		object.onRemove();
		if (alsoDelete) delete &object;
	}
}

Object* Layer::getObjectByID(const std::string& id) const
{
	ObjectMapType::const_iterator it = objects.find(id);
	return it == objects.end() ? 0 : it->second;
}

Level& Layer::getLevel() const
{
	return level;
}

const std::string& Layer::getID() const
{
	return id;
}

int Layer::getIndex() const
{
	return index;
}

float Layer::getZ() const
{
	return z;
}

const RGBA& Layer::getColor() const
{
	return color;
}

int Layer::getBlendMode() const
{
	return blendMode;
}

bool Layer::isVisible() const
{
	return visible;
}

void Layer::setupMatrix() const
{
	const Vec2f levelSize(level.getSize());
	const Vec2f scale(1.0f + z / (levelSize.x * 0.5f / app.getViewportAspect()), 1.0f + z / (levelSize.y * 0.5f));
	glTranslatef(levelSize.x * 0.5f, levelSize.y * 0.5f, z);
	glScalef(scale.x, scale.y, 1.0f);
	glTranslatef(levelSize.x * -0.5f, levelSize.y * -0.5f, 0.0f);
}

void Layer::setupBlending() const
{
	// Blending setzen
	switch(blendMode)
	{
	case 0: app.setBlending(false); break;
	case 1: app.setBlending(true); break;
	case 2: app.setBlending(true, GL_ONE, GL_ONE); break;
	case 3: app.setBlending(true, GL_SRC_ALPHA, GL_ONE); break;
	case 4: app.setBlending(true, GL_DST_COLOR, GL_ZERO); break;
	}
}

void Layer::updateObjects()
{
	updatingObjects = true;

	// alle Objekte aktualisieren
	for(ObjectMapType::iterator it = objects.begin();
		it != objects.end();
		++it)
	{
		if(it->second->isAlive())
		{
			// aktualisieren
			it->second->update();
		}
		else
		{
			// lschen
			removeObject(*it->second);
		}
	}

	updatingObjects = false;

	// neue Objekte hinzufgen
	if(!objectsToAdd.empty())
	{
		for(ObjectListType::const_iterator it = objectsToAdd.begin();
			it != objectsToAdd.end();
			++it)
		{
			addObject(*(*it).p_obj);
		}

		objectsToAdd.clear();
	}

	// Objekte entfernen
	if(!objectsToRemove.empty())
	{
		for(ObjectListType::const_iterator it = objectsToRemove.begin();
			it != objectsToRemove.end();
			++it)
		{
			removeObject(*(*it).p_obj, (*it).alsoDelete);
		}

		objectsToRemove.clear();
	}
}

void Layer::renderObjects(Object* p_forceRenderingOnTop)
{
	double modelViewMatrix[16];
	double projectionMatrix[16];
	GLint viewport[4];
	glGetDoublev(GL_MODELVIEW_MATRIX, modelViewMatrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projectionMatrix);
	glGetIntegerv(GL_VIEWPORT, viewport);
	const double left = viewport[0];
	const double right = viewport[0] + viewport[2];
	const double bottom = viewport[1];
	const double top = viewport[1] + viewport[3];

	// Wie gro ist eine Einheit auf dem Bildschirm?
	double coords[6];
	gluProject(0.0, 0.0, 0.0, modelViewMatrix, projectionMatrix, viewport, coords, coords + 1, coords + 2);
	gluProject(1.0, 1.0, 0.0, modelViewMatrix, projectionMatrix, viewport, coords + 3, coords + 4, coords + 5);
	const double sx = fabs(coords[0] - coords[3]);
	const double sy = fabs(coords[1] - coords[4]);

	// alle Objekte rendern
	bool renderOnTop = false;
	for(ObjectMapType::const_iterator it = objects.begin();
		it != objects.end();
		++it)
	{
		Object& obj = *it->second;
		if (!obj.getVisible()) continue;
		if(p_forceRenderingOnTop == &obj)
		{
			renderOnTop = true;
			continue;
		}

		if(const double r = obj.getVisibilityRadius())
		{
			// Objektposition auf dem Bildschirm ermitteln
			const Vec2f pos(obj.getPosition());
			gluProject(pos.x, pos.y, 0.0, modelViewMatrix, projectionMatrix, viewport, coords, coords + 1, coords + 2);

			// rendern, wenn sichtbar
			const double rx = r * sx;
			const double ry = r * sy;
			if(coords[0] > left - rx && coords[0] < right + rx &&
			   coords[1] > bottom - ry && coords[1] < top + ry) obj.render();
		}
		else obj.render();
	}

	if(renderOnTop && p_forceRenderingOnTop)
	{
		// Objekt rendern, das immer oben sein soll
		p_forceRenderingOnTop->render();
	}
}
