/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "imagelayer.h"
#include "level.h"
#include "object.h"
#include "playerobject.h"
#include "tilemaplayer.h"

// "Vergleicher" fr Layer
// (zum Sortieren nach z-Wert)
struct LayerSorter
{
	bool operator ()(Layer* p_lhs, Layer* p_rhs)
	{
		// z-Werte vergleichen
		return p_lhs->getZ() > p_rhs->getZ();
	}
};

Level::Level(Application& app,
			 const TiXmlElement* p_xml,
			 const ParameterMap& gameParams)
	: app(app)
	, size(0, 0)
	, backgroundColor(RGBA::BLACK)
	, useBackgroundColor(false)
	, cameraPosition(0.0f, 0.0f)
	, cameraDistance(8.0f)
	, p_world(0)
	, gravity(0.0f, 0.0f)
{
	// Gre des Levels lesen
	size = fromString<Vec2i>(p_xml->Attribute("size"));

	// Hintergrundfarbe lesen
	if(const char* p_backgroundColor = p_xml->Attribute("backgroundColor")) 
	{
		backgroundColor = fromString<RGBA>(p_backgroundColor);
		useBackgroundColor = true;
	}

	// Gravitationsvektor lesen
	if(const char* p_gravity = p_xml->Attribute("gravity")) gravity = fromString<Vec2f>(p_gravity);

	// Welt erzeugen
	b2AABB worldAABB;
	const float border = 10.0f;
	worldAABB.lowerBound = b2Vec2(-border, -border);
	worldAABB.upperBound = size + Vec2f(border, border);
	p_world = new b2World(worldAABB, gravity, true);
	p_world->SetContactListener(this);

	// Dateiname der Hintergrundmusik lesen
	if(const char* p_musicFilename = p_xml->Attribute("musicFilename")) musicFilename = p_musicFilename;

	// Layer lesen
	if(const TiXmlElement* p_layersXML = p_xml->FirstChildElement("Layers"))
	{
		int index = 0;
		for(const TiXmlElement* p_layerXML = p_layersXML->FirstChildElement();
			p_layerXML;
			p_layerXML = p_layerXML->NextSiblingElement(), ++index)
		{
			Layer* p_layer = 0;

			// Typ des Layers?
			const std::string layerType(p_layerXML->Value());
			if(layerType == "ImageLayer") p_layer = new ImageLayer(*this, index, p_layerXML, gameParams);
			else if(layerType == "TileMapLayer") p_layer = new TileMapLayer(*this, index, p_layerXML, gameParams);
			else
			{
				// ungltiger Typ
				assert(false);
			}

			// Layer hinzufgen
			addLayer(*p_layer);
		}
	}

	// Kamera in die Mitte des Levels setzen
	setCameraPosition(Vec2f(size) * 0.5f);
}

Level::~Level()
{
	// Layer lschen
	for(LayerListType::const_iterator it = layers.begin();
		it != layers.end();
		++it)
	{
		delete *it;
	}

	// Physik-Welt lschen
	delete p_world;
}

void Level::update()
{
	// alle Layer aktualisieren
	for(LayerListType::const_iterator it = layers.begin();
		it != layers.end();
		++it)
	{
		(*it)->update();
	}

	// Physik aktualisieren
	p_world->Step(app.getUpdateStepSize(), 32);
}

void Level::render(Object* p_forceRenderingOnTop)
{
	if(useBackgroundColor)
	{
		// Hintergrundfarbe
		app.clearViewport(backgroundColor);
	}

	// Projektionsmatrix einstellen
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	gluPerspective(90.0, app.getViewportAspect(), 0.001, 1000.0);

	// Kamera korrigieren und dann die Matrix bauen
	correctCamera();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	gluLookAt(cameraPosition.x, cameraPosition.y, -cameraDistance, cameraPosition.x, cameraPosition.y, 0.0, 0.0, -1.0, 0.0);

	// alle sichtbaren Layer rendern
	for(LayerListType::const_iterator it = layers.begin();
		it != layers.end();
		++it)
	{
		Layer& layer = **it;
		if(layer.isVisible()) layer.render(p_forceRenderingOnTop);
	}

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}

void Level::addResourceToCache(ResourceHandle resource)
{
	resources.push_back(resource);
}

Application& Level::getApplication() const
{
	return app;
}

const Vec2i& Level::getSize() const
{
	return size;
}

const std::string& Level::getMusicFilename() const
{
	return musicFilename;
}

Layer* Level::getLayerByID(const std::string& id) const
{
	// Layer suchen
	for(LayerListType::const_iterator it = layers.begin();
		it != layers.end();
		++it)
	{
		if((*it)->getID() == id) return *it;
	}

	// nicht gefunden
	return 0;
}

Object* Level::getObjectByID(const std::string& id) const
{
	// alle Layer durchsuchen
	for(LayerListType::const_iterator it = layers.begin();
		it != layers.end();
		++it)
	{
		if(Object* p_obj = (*it)->getObjectByID(id)) return p_obj;
	}

	// nicht gefunden
	return 0;
}

PlayerObject* Level::getPlayer(int playerNo) const
{
	const char* p_objectID = 0;
	if(playerNo == 0) p_objectID = "player1";
	else if(playerNo == 1) p_objectID = "player2";
	else return 0;
	return dynamic_cast<PlayerObject*>(getObjectByID(p_objectID));
}

const Vec2f& Level::getCameraPosition() const
{
	return cameraPosition;
}

void Level::setCameraPosition(const Vec2f& cameraPosition)
{
	this->cameraPosition = cameraPosition;
}

float Level::getCameraDistance() const
{
	return cameraDistance;
}

void Level::setCameraDistance(float cameraDistance)
{
	this->cameraDistance = cameraDistance;
	correctCamera();
}

b2World& Level::getWorld() const
{
	return *p_world;
}

void Level::Add(const b2ContactPoint* p_point)
{
	Persist(p_point);
}

void Level::Persist(const b2ContactPoint* p_point)
{
	b2Body& body1 = *p_point->shape1->GetBody();
	b2Body& body2 = *p_point->shape2->GetBody();
	Object* p_object1 = reinterpret_cast<Object*>(body1.GetUserData());
	Object* p_object2 = reinterpret_cast<Object*>(body2.GetUserData());
	const Vec2f normal(p_point->normal.x, p_point->normal.y);

	if(p_object1 && p_object1->isAlive()) p_object1->onCollision(Contact(p_object2, body1, body2, *p_point->shape1, *p_point->shape2, normal, *p_point));
	if(p_object2 && p_object2->isAlive()) p_object2->onCollision(Contact(p_object1, body2, body1, *p_point->shape2, *p_point->shape1, -normal, *p_point));
}

void Level::addLayer(Layer& layer)
{
	// Layer hinzufgen
	layers.push_back(&layer);

	// Layer anhand ihrer z-Werte sortieren
	layers.sort(LayerSorter());

	// Shapes des Layers zum Ground-Body hinzufgen
	layer.createShapes(*p_world->GetGroundBody());
}

void Level::correctCamera()
{
	if(!layers.empty())
	{
		// Die Kamera soll durch keinen der Layer fliegen,
		// und der Layer bei z = 0 soll hchstens komplett sichtbar sein.
		cameraDistance = std::max(cameraDistance, -layers.back()->getZ() + 1.0f);
		cameraDistance = std::min(cameraDistance, 0.5f * size.y);
		cameraDistance = std::min(cameraDistance, 0.5f * size.x / app.getViewportAspect());
	}

	// verhindern, dass die Kamera aus dem Level herausfliegt
	const Vec2f size(this->size);
	cameraPosition.x = std::max(cameraPosition.x, cameraDistance * app.getViewportAspect());
	cameraPosition.x = std::min(cameraPosition.x, size.x - cameraDistance * app.getViewportAspect());
	cameraPosition.y = std::max(cameraPosition.y, cameraDistance);
	cameraPosition.y = std::min(cameraPosition.y, size.y - cameraDistance);
}
