/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "layer.h"
#include "level.h"
#include "object.h"

Object::Object(Layer& layer,
			   const TiXmlElement* p_xml,
			   const ParameterMap& gameParams)
	: p_layer(&layer)
	, level(p_layer->getLevel())
	, app(p_layer->getLevel().getApplication())
	, alive(true)
	, position(0.0f, 0.0f)
	, angle(0.0f)
	, visibilityRadius(0.0f)
	, gameParams(gameParams)
	, visible(true)
{
	// ID lesen oder automatisch generieren
	if(const char* p_id = p_xml->Attribute("id")) id = p_id;
	else id = autoIDString<Object>("Object#");
}

Object::~Object()
{
	destroyBodiesAndJoints();
}

void Object::onAdd()
{

}

void Object::kill()
{
	alive = false;
}

Layer& Object::getLayer() const
{
	return *p_layer;
}

void Object::setLayer(Layer& layer)
{
	if(&layer == p_layer) return;

	// aus dem alten Layer entfernen
	p_layer->removeObject(*this, false);

	// zum neuen Layer hinzufgen
	p_layer = &layer;
	p_layer->addObject(*this);
}

const std::string& Object::getID() const
{
	return id;
}

bool Object::isAlive() const
{
	return alive;
}

b2Body& Object::getBody(size_t index) const
{
	return *bodies[index];
}

b2Joint& Object::getJoint(size_t index) const
{
	return *joints[index];
}

const Vec2f Object::getPosition() const
{
	if(bodies.empty()) return position;
	else
	{
		const b2Vec2& position = bodies.front()->GetPosition();
		return Vec2f(position.x, position.y);
	}
}

void Object::setPosition(const Vec2f& position)
{
	setTransform(position, getAngle());
}

float Object::getAngle() const
{
	return bodies.empty() ? angle : bodies.front()->GetAngle();
}

void Object::setAngle(float angle)
{
	setTransform(getPosition(), angle);
}

const Vec2f Object::getXAxis() const
{
	return bodies.empty() ? b2Vec2(1.0f, 0.0f) : bodies.front()->GetWorldVector(b2Vec2(1.0f, 0.0f));
}

const Vec2f Object::getYAxis() const
{
	return bodies.empty() ? b2Vec2(0.0f, 1.0f) : bodies.front()->GetWorldVector(b2Vec2(0.0f, 1.0f));
}

void Object::setTransform(const Vec2f& position,
						  float angle)
{
	if(bodies.empty())
	{
		this->position = position;
		this->angle = angle;
	}
	else
	{
		// Die Transformation bezieht sich immer auf den ersten (primren) Body.
		// Alle eventuell vorhandenen weiteren Bodies werden relativ dazu bewegt.
		b2Body& primaryBody = getBody(0);
		const float rotation = angle - primaryBody.GetAngle();
		const b2Mat22 rotationMatrix(rotation);

		// die Transformation auf die brigen Bodies anwenden
		for(BodyVectorType::const_iterator it = bodies.begin() + 1;
			it != bodies.end();
			++it)
		{
			const b2Vec2& currentPosition = (*it)->GetPosition();
			const float currentAngle = (*it)->GetAngle();
			(*it)->SetXForm(position + b2Mul(rotationMatrix, currentPosition - primaryBody.GetPosition()), currentAngle + rotation);
		}

		// primren Body transformieren
		primaryBody.SetXForm(position, angle);
	}
}

float Object::getVisibilityRadius() const
{
	return visibilityRadius;
}

void Object::setVisible(bool visible)
{
	this->visible = visible;
}

bool Object::getVisible() const
{
	return visible;
}

GameMode Object::getGameMode() const
{
	return gameParams.get<GameMode>("gameMode");
}

void Object::finishInit(const TiXmlElement* p_xml)
{
	// Position lesen
	Vec2f position(0.0f, 0.0f);
	if(const char* p_position = p_xml->Attribute("position")) position = fromString<Vec2f>(p_position);

	// Winkel lesen
	float angle = 0.0f;
	p_xml->QueryFloatAttribute("angle", &angle);
	angle = deg2Rad(angle);
	
	// Rendern lesen
	if (p_xml->Attribute("visible")) setVisible(fromString<bool>(p_xml->Attribute("visible")));

	// Kollisions-Bits setzen und Transformation setzen
	setupCollisionBits();
	setTransform(position, angle);
}

void Object::setupCollisionBits()
{
	// Kollisions-Bits aller Shapes setzen
	for(BodyVectorType::const_iterator it = bodies.begin();
		it != bodies.end();
		++it)
	{
		for(b2Shape* p_shape = (*it)->GetShapeList();
			p_shape;
			p_shape = p_shape->GetNext())
		{
			b2FilterData fd(p_shape->GetFilterData());
			fd.categoryBits = fd.maskBits = 1 << p_layer->getIndex();
			p_shape->SetFilterData(fd);
			level.getWorld().Refilter(p_shape);
		}
	}
}

b2Body& Object::createBody(const b2BodyDef& bodyDef)
{
	b2Body* p_body = level.getWorld().CreateBody(&bodyDef);
	assert(p_body);
	p_body->SetUserData(this);
	bodies.push_back(p_body);
	return *p_body;
}

b2Joint& Object::createJoint(const b2JointDef& jointDef)
{
	b2Joint* p_joint = level.getWorld().CreateJoint(&jointDef);
	assert(p_joint);
	p_joint->SetUserData(this);
	joints.push_back(p_joint);
	return *p_joint;
}

void Object::destroyBodiesAndJoints()
{
	// Joints zerstren
	for(JointVectorType::const_iterator it = joints.begin();
		it != joints.end();
		++it)
	{
		level.getWorld().DestroyJoint(*it);
	}

	// Bodies zerstren
	for(BodyVectorType::const_iterator it = bodies.begin();
		it != bodies.end();
		++it)
	{
		level.getWorld().DestroyBody(*it);
	}

	joints.clear();
	bodies.clear();
}

void Object::setBodyTransform(size_t bodyIndex) const
{
	// Matrix zusammenbauen
	const b2XForm& transform(getBody(bodyIndex).GetXForm());
	float matrix[16] = { transform.R.col1.x, transform.R.col1.y, 0.0f, 0.0f,
						 transform.R.col2.x, transform.R.col2.y, 0.0f, 0.0f,
						 0.0f, 0.0f, 1.0f, 0.0f,
						 transform.position.x, transform.position.y, 0.0f, 1.0f };

	// Matrix einmultiplizieren
	glMultMatrixf(matrix);
}

void Object::drawBodyQuad(size_t bodyIndex,
						  const Vec2i& topLeftOnTexture,
						  const Vec2i& sizeOnTexture,
						  const Vec2f& size,
						  bool mirrorX,
						  bool mirrorY,
						  const RGBA& color) const
{
	// Matrix setzen und Rechteck zeichnen
	glPushMatrix();
	setBodyTransform(bodyIndex);
	if(mirrorX || mirrorY) glScalef(mirrorX ? -1.0f : 1.0f, mirrorY ? -1.0f : 1.0f, 1.0f);
	glBegin(GL_QUADS);
	glColor4fv(color);
	glTexCoord2i(topLeftOnTexture.x, topLeftOnTexture.y);
	glVertex2f(size.x * -0.5f, size.y * -0.5f);
	glTexCoord2i(topLeftOnTexture.x + sizeOnTexture.x, topLeftOnTexture.y);
	glVertex2f(size.x * 0.5f, size.y * -0.5f);
	glTexCoord2i(topLeftOnTexture.x + sizeOnTexture.x, topLeftOnTexture.y + sizeOnTexture.y);
	glVertex2f(size.x * 0.5f, size.y * 0.5f);
	glTexCoord2i(topLeftOnTexture.x, topLeftOnTexture.y + sizeOnTexture.y);
	glVertex2f(size.x * -0.5f, size.y * 0.5f);
	glEnd();
	glPopMatrix();
}
