/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#ifndef OBJECT_H_INCLUDED
#define OBJECT_H_INCLUDED

#include "gamestate.h"

class Application;
class Layer;
class Level;
struct Contact;

// Basisklasse fr Spielobjekte
class Object
	: NonCopyable
{
public:
	// Konstruktor: initialisiert das Objekt
	Object(Layer& layer, const TiXmlElement* p_xml, const ParameterMap& gameParams);

	// virtueller Destruktor
	virtual ~Object();

	// aktualisiert das Objekt
	virtual void update() { }

	// rendert das Objekt
	virtual void render() { }

	// wird aufgerufen, wenn das Objekt zu einem Layer hinzugefgt wird
	virtual void onAdd();

	// wird aufgerufen, wenn das Objekt von einem Layer entfernt wird
	virtual void onRemove() { }

	// wird aufgerufen, wenn das Objekt kollidiert
	virtual void onCollision(const Contact& contact) { }

	// sorgt dafr, dass das Objekt beim nchsten Update zerstrt wird
	void kill();

	// liefert den Layer des Objekts
	// (der neue Layer bernimmt die Verantwortung fr das Objekt)
	Layer& getLayer() const;

	// setzt den Layer des Objekts
	void setLayer(Layer& layer);

	// liefert die ID des Objekts
	const std::string& getID() const;

	// prft, ob das Objekt noch lebt
	bool isAlive() const;

	// liefert den n-ten Body des Objekts
	b2Body& getBody(size_t index) const;

	// liefert den n-ten Joint des Objekts
	b2Joint& getJoint(size_t index) const;

	// liefert die Position des Objekts
	const Vec2f getPosition() const;

	// setzt die Position des Objekts
	void setPosition(const Vec2f& position);

	// liefert den Winkel des Objekts
	float getAngle() const;

	// setzt den Winkel des Objekts
	void setAngle(float angle);

	// liefert die x-Achse des Objekts
	const Vec2f getXAxis() const;

	// liefert die y-Achse des Objekts
	const Vec2f getYAxis() const;

	// setzt die Transformation des Objekts
	void setTransform(const Vec2f& position, float angle);

	// liefert den Radius fr den Sichtbarkeitstest
	float getVisibilityRadius() const;
	
	// setzt ob das Objekt gerendert werden soll
	void setVisible(bool visible);
	
	// fragt, ob das Objekt gerendert werden soll
	bool getVisible() const;

	// liefert den Spielmodus
	GameMode getGameMode() const;

protected:
	typedef std::vector<b2Body*> BodyVectorType;
	typedef std::vector<b2Joint*> JointVectorType;

	// Initialisierung fertigstellen
	void finishInit(const TiXmlElement* p_xml);

	// setzt die Kollisions-Bits
	void setupCollisionBits();

	// erzeugt einen neuen Body fr dieses Objekt
	b2Body& createBody(const b2BodyDef& bodyDef);

	// erzeugt einen neuen Joint fr dieses Objekt
	b2Joint& createJoint(const b2JointDef& jointDef);

	// zerstrt alle Bodies und Joints
	void destroyBodiesAndJoints();

	// setzt die Transformation fr einen Body
	void setBodyTransform(size_t bodyIndex) const;

	// zeichnet ein texturiertes Rechteck fr einen Body
	void drawBodyQuad(size_t bodyIndex, const Vec2i& topLeftOnTexture, const Vec2i& sizeOnTexture, const Vec2f& size, bool mirrorX = false, bool mirrorY = false, const RGBA& color = RGBA::WHITE) const;

	Layer* p_layer;					// Layer, zu dem das Objekt gehrt
	Level& level;					// der Level
	Application& app;				// das Anwendungsobjekt
	std::string id;					// ID des Objekts
	bool alive;						// Lebt das Objekt noch?
	BodyVectorType bodies;			// Krper fr die Physik
	JointVectorType joints;			// Joints fr die Physik
	Vec2f position;					// Position (falls keine Physik vorhanden)
	float angle;					// Winkel (falls keine Physik vorhanden)
	float visibilityRadius;			// Radius fr die Sichtbarkeitsbestimmung (0: kein Test)
	bool visible;					// Objekt rendern
	const ParameterMap& gameParams;	// Spielparameter
};

// Datenstruktur fr Kontakte (Kollisionen)
struct Contact
{
	Contact(Object* p_otherObject,
			b2Body& yourBody,
			b2Body& otherBody,
			b2Shape& yourShape,
			b2Shape& otherShape,
			const Vec2f& normal,
			const b2ContactPoint& contactPoint)
		: p_otherObject(p_otherObject)
		, yourBody(yourBody)
		, otherBody(otherBody)
		, yourShape(yourShape)
		, otherShape(otherShape)
		, normal(normal)
		, contactPoint(contactPoint)
	{

	}

	Object* p_otherObject;
	b2Body& yourBody;
	b2Body& otherBody;
	b2Shape& yourShape;
	b2Shape& otherShape;
	const Vec2f normal;
	const b2ContactPoint& contactPoint;
};

#endif
