/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "physicsstate.h"
#include "transitions.h"

PhysicsState::PhysicsState(Application& app)
	: State("Physics", app)
{
	// Schriftart laden
	font = app.getFontManager().getResourceHandle("fonts/arial.fnt");
}

void PhysicsState::onEnter(const ParameterMap& params)
{
	bodyIndex = 0;
	numBodies = 0;

	memset(bodies, 0, PHYSICSSTATE_MAXBODIES * sizeof(b2Body*));

	// Box 2D
	worldAABB.lowerBound.Set(-100.0f, -100.0f);
	worldAABB.upperBound.Set(100.0f, 100.0f);

	b2Vec2 gravity(0.0f, 10.0f);
	bool doSleep = true;

	world = new b2World(worldAABB, gravity, doSleep);

	// Boden
	b2BodyDef groundBodyDef;
	groundBodyDef.position.Set(16.0f, 25.0f);
	
	b2Body* ground = world->CreateBody(&groundBodyDef);
	
	b2PolygonDef groundShapeDef;
	groundShapeDef.restitution = 0.75f; // "hpfen" erlauben
	groundShapeDef.friction = 0.85f;
	groundShapeDef.SetAsBox(16.0f, 1.0f);
	
	ground->CreateShape(&groundShapeDef);

	// linke Wand
	groundShapeDef.SetAsBox(1.0f, 12.0f, b2Vec2(-17.0f, -13.0f), 0.0f);
	ground->CreateShape(&groundShapeDef);

	// rechte Wand
	groundShapeDef.SetAsBox(1.0f, 12.0f, b2Vec2(17.0f, -13.0f), 0.0f);
	ground->CreateShape(&groundShapeDef);
	
	// Bodies mit Joints
	
	// body 1: Rechteck
	b2BodyDef bodyDef;
	bodyDef.position.Set(8.0f, 4.0f);
	bodyDef.allowSleep = false;
	
	// Schwerpunkt verschieben
	bodyDef.massData.mass = 1000.0f;
	b2Vec2 center(0.5f, 2.5f);
	bodyDef.massData.center = center;
	bodyDef.massData.I = 5.0f;
	body1 = world->CreateBody(&bodyDef);
	
	b2PolygonDef body1Shape;
	body1Shape.restitution = 0.75f;
	body1Shape.density = 1.0f;
	body1Shape.friction = 0.0f;
	body1Shape.SetAsBox(2.0f, 1.0f);
	body1->CreateShape(&body1Shape);
	
	// Rder
	b2CircleDef wheel;
	wheel.density = 40.0f;
	wheel.friction = 0.9f;
	wheel.radius = 1.0f;
	
	// Rad 1
	bodyDef.position.Set(6.5f, 4.5f);
	body2 = world->CreateBody(&bodyDef);
	body2->CreateShape(&wheel);
	body2->SetMassFromShapes();
	
	// Rad 2
	bodyDef.position.Set(9.5f, 4.5f);
	body3 = world->CreateBody(&bodyDef);
	body3->CreateShape(&wheel);
	body3->SetMassFromShapes();
	
	// Joints
	b2RevoluteJointDef rjd;
	rjd.Initialize(body1, body2, body2->GetWorldCenter());
	rjd.collideConnected = false;
	rjd.maxMotorTorque = 5000.0f;
	rjd.motorSpeed = 0.0f;
	rjd.enableMotor = true;
	rj = (b2RevoluteJoint*) world->CreateJoint(&rjd);
	
	rjd.Initialize(body1, body3, body3->GetWorldCenter());
	rjd.collideConnected = false;
	rjd.enableMotor = false;
	world->CreateJoint(&rjd);
	
	
	// ein paar Boxen erzeugen
	for(int i = 0; i < 8; ++i) {
		addBody(2.0f + 1.75f * float(i), 1.0f + 2.25f * float(i % 4));
	}
}

void PhysicsState::onLeave()
{

}

void PhysicsState::onGotFocus()
{

}

void PhysicsState::onLostFocus()
{

}

void PhysicsState::update()
{
	if(app.wasKeyPressed(SDLK_ESCAPE) || app.wasKeyPressed(SDLK_F5))
	{
		// Zustand verlassen
		app.popState();
		app.startTransition(new CubeTransition(app, 0.75f));
		return;
	}

	if(app.isKeyDown(SDLK_F6)) {
		addBody(16.0f, 1.0f);
	}
	
	if(app.isKeyDown(SDLK_F7)) {
		removeBody();
	}
	
	rj->SetMotorSpeed(0.0f);
	
	if(app.isKeyDown(SDLK_a)) {
		rj->SetMotorSpeed(-25.0f);
	}
	
	if(app.isKeyDown(SDLK_d)) {
		rj->SetMotorSpeed(25.0f);
	}

	// Iterations sollten woanders definiert werden
	world->Step(app.getUpdateStepSize(), 10);
}

void PhysicsState::render()
{
	app.clearViewport(RGBA(0.5f, 0.5f, 1.0f, 0.0f));

	// Projektionsmatrix einstellen
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, 32.0, 24.0, 0.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	
	app.setTexturing(false);
	
	glLoadIdentity();
	
	// eigene Bodies malen
	glColor3f(0.0f, 1.0f, 0.0f);
	drawBody(body1);
	glColor3f(0.0f, 0.0f, 1.0f);
	drawBody(body2);
	drawBody(body3);

	// Fr die Farbberechnung
	int sep = (PHYSICSSTATE_MAXBODIES / 3) + 15;
	float norm = 1.0f / float(sep);

	for(int i = 0; i < PHYSICSSTATE_MAXBODIES; ++i) {
		if(bodies[i] == 0) {
			continue;
		}
		// schwarz, rot, gold
		glColor3f(
			norm * float(std::min(sep, i)),
			norm * float(std::min(sep, std::max(0, i - sep))),
			norm * float(std::min(sep, std::max(0, i - sep - sep)))
		);
		drawBody(bodies[i]);
	}


	font->drawText(Vec2f(0.0f, 0.0f), 1.0f, RGBA(1.0f, 1.0f, 0.0f), "Willkommen im Physics-State!\nVerlassen mit ESC oder F5!");	

	// schlechter Code:
	char text[255];
	sprintf(text, "Boxen: %d / %d\nF6 fr neue Bodies\nF7 zum lschen!", numBodies, PHYSICSSTATE_MAXBODIES);

	font->drawText(Vec2f(24.0f, 0.0f), 1.0f, RGBA(1.0f, 0.0f, 0.0f), text);

}

void PhysicsState::drawBody(b2Body* body) {

	// Transformation fr lokale Koordinaten des Bodies in Weltkoordinaten
	const b2XForm& xf = body->GetXForm();
	// Anzahl der Vertices
	int vertexCount = 0;

	// Alle Shapes durchlaufen (sollte immer nur eine sein)
	for(b2Shape* s = body->GetShapeList(); s; s = s->GetNext()) {

		if(s->GetType() == e_circleShape) {
			b2CircleShape* circle = (b2CircleShape*) s;
			float radius = circle->GetRadius();
			b2Vec2 center = b2Mul(xf, circle->GetLocalPosition());
			float k_segments = 16.0f;
			float k_increments = 2.0f * b2_pi / k_segments;
			float theta = 0.0f;
			glBegin(GL_POLYGON);
			for(int i = 0; i < k_segments; ++i) {
				b2Vec2 v = center + radius * b2Vec2(cosf(theta), sinf(theta));
				glVertex2f(v.x, v.y);
				theta += k_increments;
			}
			glEnd();
		}

		// die Methode hier kann nur polygonshapes malen
		if(s->GetType() != e_polygonShape) {
			return;
		}

		// Vertices auslesen
		b2PolygonShape* ps = (b2PolygonShape*) s;
		vertexCount = ps->GetVertexCount();
		const b2Vec2* vertices = ps->GetVertices();
		b2Vec2 vert;

		// Alle Vertices dieser Shape zeichnen
		glBegin(GL_POLYGON);
		for(int i = 0; i < vertexCount; ++i) {
			// in Weltkoordinaten Transformieren
			vert = b2Mul(xf, vertices[i]);
			glVertex2f(vert.x, vert.y);
		}
		glEnd();
	}
}

void PhysicsState::addBody(float x, float y) {

	if(bodies[bodyIndex] != 0) {
		// mehr Boxen als erlaubt, lteste wieder lschen
		world->DestroyBody(bodies[bodyIndex]);
	}

	// Body erzeugen
	b2BodyDef bodyDef;
	bodyDef.position.Set(x, y);
	bodies[bodyIndex] = world->CreateBody(&bodyDef);

	b2ShapeDef* shapeDef;
	
	// Form erzeugen
	if(bodyIndex % 8 != 0) {
		b2CircleDef circle = b2CircleDef();
		circle.radius = 0.55f;
		shapeDef = &circle;
	} else {
		b2PolygonDef box = b2PolygonDef();
		box.SetAsBox(0.55f, 0.55f); // Gre der Boxen
		shapeDef = &box;
	}
	shapeDef->density = 1.0f;  // Die Dichte bestimmt das Gewicht
	shapeDef->friction = 0.3f;  // Reibung
	bodies[bodyIndex]->CreateShape(shapeDef);
	bodies[bodyIndex]->SetMassFromShapes();

	bodyIndex = (bodyIndex + 1) % PHYSICSSTATE_MAXBODIES; // wrap-around
	numBodies = std::min(numBodies + 1, PHYSICSSTATE_MAXBODIES); // wenn das bodies-Array voll ist, bleibt es voll
}

void PhysicsState::removeBody() {
	bodyIndex = (PHYSICSSTATE_MAXBODIES + bodyIndex - 1) % PHYSICSSTATE_MAXBODIES;
	if(bodies[bodyIndex] != 0) {
		world->DestroyBody(bodies[bodyIndex]);
		bodies[bodyIndex] = 0;
		numBodies = std::max(0, numBodies - 1);
	}
}
