/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "playerobject.h"
#include "layer.h"
#include "level.h"
#include "textobject.h"

PlayerObject::PlayerObject(Layer& layer,
						   const TiXmlElement* p_xml,
						   const ParameterMap& gameParams)
	: Object(layer, p_xml, gameParams)
	, playerNo(0)
	, color(RGBA::WHITE)
	, points(0)
	, pointsF(0.0f)
	, health(100)
	, healthF(100.0f)
	, mirrored(false)
	, motorSpeed(0.0f)
	, frontWheelCohesionForce(0.0f, 0.0f)
	, backWheelCohesionForce(0.0f, 0.0f)
	, frontWheelOnGround(false)
	, backWheelOnGround(false)
	, active(true)
	, changeLayer(false)
{
	visibilityRadius = 2.0f;

	// Spielernummer lesen
	p_xml->QueryIntAttribute("playerNo", &playerNo);

	// Blickrichtung lesen
	int direction = 1;
	p_xml->QueryIntAttribute("direction", &direction);
	mirrored = direction == -1;

	// Farbe setzen
	if(playerNo == 0) color = RGBA(242.0f / 255.0f, 89.0f / 255.0f, 58 / 255.0f);
	else if(playerNo == 1) color = RGBA(41.0f / 255.0f, 201.0f / 255.0f, 239.0f / 255.0f);

	// Spieler 2 soll nur dann existieren, wenn wir in einem 2-Spieler-Modus sind.
	if(playerNo == 1 &&
	   getGameMode() != GM_TIME_2P &&
	   getGameMode() != GM_CTG &&
	   getGameMode() != GM_KOH)
	{
		kill();
	}

	// Sprites laden
	level.addResourceToCache(sprites = app.getTextureManager().getResourceHandle("sprites/player.png"));

	// Motorsound laden
	level.addResourceToCache(motorSound = app.getSoundManager().getResourceHandle("sounds/motor.ogg"));
	motorSoundInstance = motorSound->createInstance(true, true, 0.0f);

	// Steuerungsmodus holen
	controlMode = gameParams.get<ControlMode>("controlMode" + toString(playerNo + 1));

	// Physik initialisieren
	setup();

	// Initialisierung fertigstellen
	finishInit(p_xml);
}

PlayerObject::~PlayerObject()
{
	// Sound stoppen
	motorSoundInstance->stop();
}

void PlayerObject::update()
{
	if (changeLayer)
	{
		// Layer wechseln
		Layer* layer = level.getLayerByID(targetLayer);
		if (layer) setLayer(*layer);
		changeLayer = false;
		setup();
	}

	b2Body& frameBody = getBody(0);
	b2Body& frontWheelBody = getBody(1);
	b2Body& backWheelBody = getBody(2);
	b2Body& driverBody = getBody(3);
	b2RevoluteJoint& frameFrontWheelJoint = static_cast<b2RevoluteJoint&>(getJoint(0));
	b2RevoluteJoint& frameBackWheelJoint = static_cast<b2RevoluteJoint&>(getJoint(1));
	b2RevoluteJoint& frameDriverJoint = static_cast<b2RevoluteJoint&>(getJoint(2));

	// Motor-Sound anpassen
	motorSoundInstance->setPitch(1.0f + 0.001f * motorSpeed * motorSpeed);
	motorSoundInstance->setVolume(0.1f + 0.01f * fabsf(motorSpeed));

	// Adhsionskrfte ausben und verringern
	frontWheelBody.ApplyForce(frontWheelCohesionForce, frontWheelBody.GetPosition());
	backWheelBody.ApplyForce(backWheelCohesionForce, backWheelBody.GetPosition());
	frontWheelCohesionForce *= 0.5f;
	backWheelCohesionForce *= 0.5f;

	if(controlMode == CM_JAN)
	{
		// Steuerung
		static SDLKey normalKeys[2][C_MAX_VALUE] =
		{
			SDLK_RIGHT,	SDLK_LEFT,	SDLK_UP,	SDLK_DOWN,
			SDLK_d,		SDLK_a,		SDLK_w,		SDLK_s
		};

		static SDLKey mirroredKeys[2][C_MAX_VALUE] =
		{
			SDLK_LEFT,	SDLK_RIGHT,	SDLK_DOWN,	SDLK_UP,
			SDLK_a,		SDLK_d,		SDLK_s,		SDLK_w
		};

		SDLKey* keys;

		if(mirrored) keys = mirroredKeys[playerNo];
		else keys = normalKeys[playerNo];

		// Zielwinkel fr den Fahrer bestimmen
		float targetDriverAngle = 0.0f;
		if(active)
		{
			if(app.isKeyDown(keys[C_LEAN_FORWARD])) targetDriverAngle = 0.4f;
			if(app.isKeyDown(keys[C_LEAN_BACKWARD])) targetDriverAngle = -0.4f;
		}

		// Fahrer anpassen
		driverBody.ApplyTorque(50.0f * (targetDriverAngle - frameDriverJoint.GetJointAngle()) - 5.0f * driverBody.GetAngularVelocity());

		// Schwerpunkt des Motorrads verlagern
		b2MassData frameMass;
		frameMass.mass = 1.0f;
		frameMass.center = b2Vec2((mirrored ? 0.75f : -0.75f) + frameDriverJoint.GetJointAngle() * 7.5f, 1.25f);
		frameMass.I = 1.0f;
		frameBody.SetMass(&frameMass);

		// zustzliches Drehmonent ausben
		frameBody.ApplyTorque(frameDriverJoint.GetJointAngle() * 150.0f);

		float targetMotorSpeed = 0.0f;
		if(active && app.isKeyDown(keys[C_GO_FORWARD]))
		{
			// Gas geben
			frameBackWheelJoint.EnableMotor(true);
			frameBackWheelJoint.SetMaxMotorTorque(50.0f);
			frameBackWheelJoint.SetMotorSpeed(motorSpeed);
			backWheelBody.WakeUp();
			targetMotorSpeed = mirrored ? -40.0f : 40.0f;
		}
		else
		{
			// Motor aus
			frameBackWheelJoint.EnableMotor(false);
		}

		// Motorgeschwindigkeit der Zielgeschwindigkeit anpassen
		motorSpeed = 0.98f * motorSpeed + 0.02f * targetMotorSpeed;

		if(!active || app.isKeyDown(keys[C_GO_BACKWARD]))
		{
			// bremsen
			frameBackWheelJoint.EnableMotor(true);
			frameBackWheelJoint.SetMotorSpeed(0.0f);
			frameBackWheelJoint.SetMaxMotorTorque(25.0f);
			frameFrontWheelJoint.EnableMotor(true);
			frameFrontWheelJoint.SetMaxMotorTorque(25.0f);
		}
		else
		{
			// nicht bremsen
			frameFrontWheelJoint.EnableMotor(false);
		}

		// umdrehen
		if(active && app.isKeyDown(keys[C_GO_BACKWARD])
			&& frameBackWheelJoint.GetJointSpeed() > -0.25f
			&& frameBackWheelJoint.GetJointSpeed() < 0.25f)
		{
			mirrored = !mirrored;
			setup();
			frontWheelOnGround = backWheelOnGround = 0;
		}
	}
	else if(controlMode == CM_WILLI)
	{
		static const SDLKey keys[A_MAX_VALUE][2] = { SDLK_UP,		SDLK_w,
													 SDLK_DOWN,		SDLK_s,
													 SDLK_RIGHT,	SDLK_d,
													 SDLK_LEFT,		SDLK_a,
													 SDLK_RCTRL,	SDLK_SPACE };

		// Zielwinkel fr den Fahrer bestimmen
		float targetDriverAngle = 0.0f;
		if(active)
		{
			if(app.isKeyDown(keys[A_LEAN_FORWARD][playerNo])) targetDriverAngle = 0.4f;
			if(app.isKeyDown(keys[A_LEAN_BACK][playerNo])) targetDriverAngle = -0.4f;
		}

		// Fahrer anpassen
		driverBody.ApplyTorque(50.0f * (targetDriverAngle - frameDriverJoint.GetJointAngle()) - 5.0f * driverBody.GetAngularVelocity());

		// Schwerpunkt des Motorrads verlagern
		b2MassData frameMass;
		frameMass.mass = 1.0f;
		frameMass.center = b2Vec2((mirrored ? 0.75f : -0.75f) + frameDriverJoint.GetJointAngle() * 7.5f, 1.25f);
		frameMass.I = 1.0f;
		frameBody.SetMass(&frameMass);

		// zustzliches Drehmonent ausben
		frameBody.ApplyTorque(frameDriverJoint.GetJointAngle() * 150.0f);

		float targetMotorSpeed = 0.0f;
		if(active && app.isKeyDown(keys[A_ACCELERATE][playerNo]))
		{
			// Gas geben
			frameBackWheelJoint.EnableMotor(true);
			frameBackWheelJoint.SetMaxMotorTorque(50.0f);
			frameBackWheelJoint.SetMotorSpeed(motorSpeed);
			backWheelBody.WakeUp();
			targetMotorSpeed = mirrored ? -40.0f : 40.0f;
		}
		else
		{
			// Motor aus
			frameBackWheelJoint.EnableMotor(false);
		}

		// Motorgeschwindigkeit der Zielgeschwindigkeit anpassen
		motorSpeed = 0.98f * motorSpeed + 0.02f * targetMotorSpeed;

		if(active && app.isKeyDown(keys[A_BRAKE][playerNo]))
		{
			// bremsen
			frameBackWheelJoint.EnableMotor(true);
			frameBackWheelJoint.SetMotorSpeed(0.0f);
			frameBackWheelJoint.SetMaxMotorTorque(25.0f);
			frameFrontWheelJoint.EnableMotor(true);
			frameFrontWheelJoint.SetMaxMotorTorque(25.0f);
		}
		else
		{
			// nicht bremsen
			frameFrontWheelJoint.EnableMotor(false);
		}

		if(active && app.wasKeyReallyPressed(keys[A_TURN][playerNo]))
		{
	//		if(frontWheelOnGround && backWheelOnGround &&
	//		   frameBody.GetLinearVelocity().LengthSquared() < 0.5f)
			{
				// umdrehen
				mirrored = !mirrored;
				setup();
				frontWheelOnGround = backWheelOnGround = 0;
			}
		}
	}

}

void PlayerObject::render()
{
	app.setBlending(true);
	app.setTexturing(true);
	sprites->bind();

	// Rahmen, Rder und Fahrer rendern
	const int y = playerNo * 256;
	drawBodyQuad(3, Vec2i(512, y), Vec2i(256, 256), Vec2f(1.6f, 1.6f), mirrored);
	drawBodyQuad(0, Vec2i(0, y), Vec2i(384, 256), Vec2f(3.0f, 2.0f), mirrored);
	drawBodyQuad(1, Vec2i(384, y), Vec2i(128, 128), Vec2f(1.1f, 1.1f));
	drawBodyQuad(2, Vec2i(384, y), Vec2i(128, 128), Vec2f(1.1f, 1.1f));
}

void PlayerObject::onCollision(const Contact& contact)
{
	if(contact.otherBody.IsStatic() &&
	   !contact.otherShape.IsSensor())
	{
		b2Body& frontWheelBody = getBody(1);
		b2Body& backWheelBody = getBody(2);
		b2Body& driverBody = getBody(3);

		if(&contact.yourBody == &frontWheelBody)
		{
			frontWheelCohesionForce = contact.normal * 0.25f * frontWheelBody.GetLinearVelocity().LengthSquared();
			frontWheelOnGround = 10;
		}
		else if(&contact.yourBody == &backWheelBody)
		{
			backWheelCohesionForce = contact.normal * 0.1f * backWheelBody.GetLinearVelocity().LengthSquared();
			backWheelOnGround = 10;
		}
		else if(&contact.yourBody == &driverBody)
		{
			// float damage = std::min(2.0f, contact.contactPoint.velocity.Length() * 0.5f);
			// reduceHealth(static_cast<int>(damage));
		}
	}
}

void PlayerObject::setup()
{
	const Vec2f position(getPosition());
	const float angle = getAngle();

	// Geschwindigkeiten speichern
	std::vector<b2Vec2> linearVelocities;
	std::vector<float> angularVelocities;
	if(!bodies.empty())
	{
		for(BodyVectorType::const_iterator it = bodies.begin();
			it != bodies.end();
			++it)
		{
			linearVelocities.push_back((*it)->GetLinearVelocity());
			angularVelocities.push_back((*it)->GetAngularVelocity());
		}
	}

	destroyBodiesAndJoints();
	const float x = mirrored ? -1.0f : 1.0f;

	// Rahmen erzeugen
	b2BodyDef frameBodyDef;
	frameBodyDef.linearDamping = 0.1f;
	frameBodyDef.angularDamping = 1.0f;
	frameBodyDef.massData.mass = 1.0f;
	frameBodyDef.massData.center = b2Vec2(0.25f * x, 1.25f);
	frameBodyDef.massData.I = 1.0f;
	b2Body& frameBody = createBody(frameBodyDef);
	b2PolygonDef framePolygonDef;
	framePolygonDef.SetAsBox(1.0f, 0.5f);
	if(getGameMode() == GM_TIME_2P) framePolygonDef.filter.groupIndex = -1;
	frameBody.CreateShape(&framePolygonDef);

	// Vorderrad erzeugen
	b2BodyDef frontWheelBodyDef;
	frontWheelBodyDef.angularDamping = 0.25f;
	b2Body& frontWheelBody = createBody(frontWheelBodyDef);
	b2CircleDef frontWheelCircleDef;
	frontWheelCircleDef.radius = 0.5f;
	frontWheelCircleDef.density = 1.0f;
	frontWheelCircleDef.restitution = 0.1f;
	frontWheelCircleDef.friction = 5.0f;
	if(getGameMode() == GM_TIME_2P) frontWheelCircleDef.filter.groupIndex = -1;
	frontWheelBody.CreateShape(&frontWheelCircleDef);
	frontWheelBody.SetMassFromShapes();
	frontWheelBody.SetXForm(b2Vec2(1.0f * x, 0.5f), 0.0f);

	// Rahmen und Vorderrad verbinden
	b2RevoluteJointDef frameFrontWheelJointDef;
	frameFrontWheelJointDef.Initialize(&frameBody, &frontWheelBody, b2Vec2(1.0f * x, 0.5f));
	createJoint(frameFrontWheelJointDef);

	// Hinterrad erzeugen
	b2BodyDef backWheelBodyDef;
	backWheelBodyDef.angularDamping = 0.25f;
	b2Body& backWheelBody = createBody(backWheelBodyDef);
	b2CircleDef backWheelCircleDef;
	backWheelCircleDef.radius = 0.5f;
	backWheelCircleDef.density = 1.0f;
	backWheelCircleDef.restitution = 0.1f;
	backWheelCircleDef.friction = 5.0f;
	if(getGameMode() == GM_TIME_2P) backWheelCircleDef.filter.groupIndex = -1;
	backWheelBody.CreateShape(&backWheelCircleDef);
	backWheelBody.SetMassFromShapes();
	backWheelBody.SetXForm(b2Vec2(-1.05f * x, 0.48f), 0.0f);

	// Rahmen und Hinterrad verbinden
	b2RevoluteJointDef frameBackWheelJointDef;
	frameBackWheelJointDef.Initialize(&frameBody, &backWheelBody, b2Vec2(-1.05f * x, 0.48f));
	createJoint(frameBackWheelJointDef);

	// Fahrer erzeugen
	b2BodyDef driverBodyDef;
	b2Body& driverBody = createBody(driverBodyDef);
	b2PolygonDef driverPolygonDef;
	driverPolygonDef.SetAsBox(0.25f, 0.5f);
	driverPolygonDef.density = 0.1f;
	if(getGameMode() == GM_TIME_2P) driverPolygonDef.filter.groupIndex = -1;
	driverBody.CreateShape(&driverPolygonDef);
	driverBody.SetMassFromShapes();
	driverBody.SetXForm(b2Vec2(-0.4f * x, -1.0f), 0.0f);

	// Rahmen und Fahrer verbinden
	b2RevoluteJointDef frameDriverJointDef;
	frameDriverJointDef.Initialize(&frameBody, &driverBody, b2Vec2(-0.4f * x, 0.0f));
	frameDriverJointDef.lowerAngle = -0.4f;
	frameDriverJointDef.upperAngle = 0.4f;
	frameDriverJointDef.enableLimit = true;
	createJoint(frameDriverJointDef);

	// Transformation wiederherstellen und Kollisions-Bits setzen
	setTransform(position, angle);
	setupCollisionBits();

	// Geschwindigkeiten wiederherstellen
	for(size_t i = 0; i < linearVelocities.size(); ++i)
	{
		bodies[i]->SetLinearVelocity(linearVelocities[i]);
		bodies[i]->SetAngularVelocity(angularVelocities[i]);
	}
}

int PlayerObject::getPoints() const
{
	return points;
}

void PlayerObject::addPoints(int amount)
{
	if(!amount || !active) return;

	points += amount;

	// Textobjekt erzeugen
	TiXmlElement xml("TextObject");
	MTRand& rng = MTRand::globalMTRand;
	xml.SetAttribute("text", (std::string(amount > 0 ? "+" : "") + toString(amount)).c_str());
	xml.SetAttribute("size", "1.5");
	xml.SetAttribute("position", toString(getPosition() + Vec2f(rng.randomFloat(-0.5f, 0.5f), -1.5f)).c_str());
	xml.SetAttribute("color", toString(amount > 0 ? RGBA::GREEN : RGBA::RED).c_str());
	xml.SetAttribute("velocity", toString(Vec2f(rng.randomFloat(-0.1f, 0.1f), rng.randomFloat(-0.9f, -1.1f))).c_str());
	xml.SetAttribute("timeToLive", "1.75");
	p_layer->addObject(*new TextObject(*p_layer, &xml, gameParams));
}

void PlayerObject::addPoints(float amount)
{
	pointsF += amount;
	const int wholePoints = static_cast<int>(pointsF);
	if(wholePoints)
	{
		addPoints(wholePoints);
		pointsF -= wholePoints;
	}
}

void PlayerObject::reduceHealth(int amount)
{
	if(!amount || !active) return;

	health = std::min(100, std::max(0, health - amount));

	if(health == 0)
	{
		// Spieler ist tot
		deactivate();

		// Textobjekt erzeugen
		TiXmlElement xml("TextObject");
		MTRand& rng = MTRand::globalMTRand;
		xml.SetAttribute("text", "TOT");
		xml.SetAttribute("position", toString(getPosition() + Vec2f(rng.randomFloat(-0.5f, 0.5f), -1.5f)).c_str());
		xml.SetAttribute("color", toString(RGBA::RED).c_str());
		xml.SetAttribute("velocity", toString(Vec2f(rng.randomFloat(-0.1f, 0.1f), rng.randomFloat(-0.9f, -1.1f))).c_str());
		xml.SetAttribute("timeToLive", toString(10.0f).c_str());
		p_layer->addObject(*new TextObject(*p_layer, &xml, gameParams));
	}
}

void PlayerObject::reduceHealth(float amount)
{
	healthF += amount;
	const int wholeHealth = static_cast<int>(healthF);
	if(wholeHealth)
	{
		reduceHealth(wholeHealth);
		healthF -= amount;
	}
}

int PlayerObject::getHealth() const
{
	return health;
}

void PlayerObject::activate()
{
	active = true;
}

void PlayerObject::deactivate()
{
	active = false;
}

const RGBA& PlayerObject::getColor() const
{
	return color;
}

bool PlayerObject::isMirrored() const
{
	return mirrored;
}

void PlayerObject::setMirrored(bool mirrored)
{
	this->mirrored = mirrored;
	setup();
}

void PlayerObject::addVelocity( b2Vec2 velocity )
{
	b2Body& body = getBody(0);

#ifdef ACCELERATE_ONCE
	if(velocity.x >= 0)
	{
		if(body.GetLinearVelocity().x > velocity.x)
			velocity.x = body.GetLinearVelocity().x;
	}
	else
	{
		if(body.GetLinearVelocity().x < velocity.x)
			velocity.x = body.GetLinearVelocity().x;
	}

	if(velocity.y >= 0)
	{
		if(body.GetLinearVelocity().y > velocity.y)
			velocity.y = body.GetLinearVelocity().y;		
	}
	else
	{
		if(body.GetLinearVelocity().y < velocity.y)
			velocity.y = body.GetLinearVelocity().y;
	}
	body.SetLinearVelocity(velocity);
#endif

#ifndef ACCELERATE_ONCE
	body.SetLinearVelocity(body.GetLinearVelocity() + velocity);
#endif
}

void PlayerObject::setLayerDelayed(std::string layerID)
{
	targetLayer = layerID;
	changeLayer = true;
}
