/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#ifndef COLOR_H_INCLUDED
#define COLOR_H_INCLUDED

// RGBA-Farbe, dargestellt als 4 floats
struct RGBA
{
	// Standardkonstruktor: initialisiert nichts
	RGBA();

	// initialisiert alle Farbkanle mit demselben Wert und separat dazu Alpha
	explicit RGBA(float rgb, float a = 1.0f);

	// initialisiert den Vektor mit r, g, b und a
	RGBA(float r, float g, float b, float a = 1.0f);

	// initialisiert den Vektor mit p_rgba[0], p_rgba[1], p_rgba[2] und p_rgba[3]
	explicit RGBA(const float* p_rgba);

	// Konvertierung nach 32-Bit-RGBA
	Uint32 toRGBA32() const;

	// setzt den Alphawert und liefert die Farbe
	RGBA& setAlpha(float alpha);

	// Casting-Operator in konstanten Zeigertyp
	operator const float*() const;

	// Casting-Operator in Zeigertyp
	operator float*();

	// Addition und Zuweisung
	RGBA& operator +=(const RGBA& rhs);

	// Subtraktion und Zuweisung
	RGBA& operator -=(const RGBA& rhs);

	// Multiplikation mit Skalar und Zuweisung
	RGBA& operator *=(float rhs);

	// komponentenweise Multiplikation mit Vektor und Zuweisung
	RGBA& operator *=(const RGBA& rhs);

	// Division durch Skalar und Zuweisung
	RGBA& operator /=(float rhs);

	// komponentenweise Division durch Vektor und Zuweisung
	RGBA& operator /=(const RGBA& rhs);

	// Negation
	RGBA operator -() const;

	// Initialisierung aus 32-Bit-RGBA
	static const RGBA fromRGBA32(Uint32 rgba32);

	float r;	// Rot-Komponente
	float g;	// Grn-Komponente
	float b;	// Blau-Komponente
	float a;	// Alpha-Komponente (Opazitt)

	static const RGBA TRANSPARENT_BLACK;	// transparentes Schwarz
	static const RGBA BLACK;				// Schwarz
	static const RGBA WHITE;				// Wei
	static const RGBA RED;					// Rot
	static const RGBA GREEN;				// Grn
	static const RGBA BLUE;					// Blau
	static const RGBA YELLOW;				// Gelb
	static const RGBA CYAN;					// Zyan
	static const RGBA MAGENTA;				// Magenta
};

inline RGBA::RGBA()
{

}

inline RGBA::RGBA(float rgb,
				  float a)
	: r(rgb)
	, g(rgb)
	, b(rgb)
	, a(a)
{

}

inline RGBA::RGBA(float r,
				  float g,
				  float b,
				  float a)
	: r(r)
	, g(g)
	, b(b)
	, a(a)
{

}

inline RGBA::RGBA(const float* p_rgba)
	: r(p_rgba[0])
	, g(p_rgba[1])
	, b(p_rgba[2])
	, a(p_rgba[3])
{

}

inline Uint32 RGBA::toRGBA32() const
{
	// Farbwerte auf [0, 255] abbilden
	int r8 = static_cast<int>(255.99998f * r);
	if(r8 < 0) r8 = 0; else if(r8 > 255) r8 = 255;
	int g8 = static_cast<int>(255.99998f * g);
	if(g8 < 0) g8 = 0; else if(g8 > 255) g8 = 255;
	int b8 = static_cast<int>(255.99998f * b);
	if(b8 < 0) b8 = 0; else if(b8 > 255) b8 = 255;
	int a8 = static_cast<int>(255.99998f * a);
	if(a8 < 0) a8 = 0; else if(a8 > 255) a8 = 255;

	// Bits zusammensetzen
	return r8 | (g8 << 8) | (b8 << 16) | (a8 << 24);
}

inline RGBA& RGBA::setAlpha(float alpha)
{
	a = alpha;
	return *this;
}

inline RGBA::operator const float*() const
{
	return &r;
}

inline RGBA::operator float*()
{
	return &r;
}

inline RGBA RGBA::operator -() const
{
	return RGBA(-r, -g, -b, -a);
}

inline RGBA& RGBA::operator +=(const RGBA& rhs)
{
	r += rhs.r;
	g += rhs.g;
	b += rhs.b;
	a += rhs.a;
	return *this;
}

inline RGBA& RGBA::operator -=(const RGBA& rhs)
{
	r -= rhs.r;
	g -= rhs.g;
	b -= rhs.b;
	a -= rhs.a;
	return *this;
}

inline RGBA& RGBA::operator *=(float rhs)
{
	r *= rhs;
	g *= rhs;
	b *= rhs;
	a *= rhs;
	return *this;
}

inline RGBA& RGBA::operator *=(const RGBA& rhs)
{
	r *= rhs.r;
	g *= rhs.g;
	b *= rhs.b;
	a *= rhs.a;
	return *this;
}

inline RGBA& RGBA::operator /=(float rhs)
{
	r /= rhs;
	g /= rhs;
	b /= rhs;
	a /= rhs;
	return *this;
}

inline RGBA& RGBA::operator /=(const RGBA& rhs)
{
	r /= rhs.r;
	g /= rhs.g;
	b /= rhs.b;
	a /= rhs.a;
	return *this;
}

// Farbaddition
inline RGBA operator +(const RGBA& lhs,
					   const RGBA& rhs)
{
	return RGBA(lhs.r + rhs.r, lhs.g + rhs.g, lhs.b + rhs.b, lhs.a + rhs.a);
}

// Farbsubtraktion
inline RGBA operator -(const RGBA& lhs,
					   const RGBA& rhs)
{
	return RGBA(lhs.r - rhs.r, lhs.g - rhs.g, lhs.b - rhs.b, lhs.a - rhs.a);
}

// Skalar von links multipliziert mit Farbe
inline RGBA operator *(float lhs,
					   const RGBA& rhs)
{
	return RGBA(lhs * rhs.r, lhs * rhs.g, lhs * rhs.b, lhs * rhs.a);
}

// Skalar von rechts multipliziert mit Farbe
inline RGBA operator *(const RGBA& lhs,
					   float rhs)
{
	return RGBA(lhs.r * rhs, lhs.g * rhs, lhs.b * rhs, lhs.a * rhs);
}

// komponentenweise Farbmultiplikation
inline RGBA operator *(const RGBA& lhs,
					   const RGBA& rhs)
{
	return RGBA(lhs.r * rhs.r, lhs.g * rhs.g, lhs.b * rhs.b, lhs.a * rhs.a);
}

// Farbe dividiert durch Skalar
inline RGBA operator /(const RGBA& lhs,
					   float rhs)
{
	return RGBA(lhs.r / rhs, lhs.g / rhs, lhs.b / rhs, lhs.a / rhs);
}

// komponentenweise Farbdivision
inline RGBA operator /(const RGBA& lhs,
					   const RGBA& rhs)
{
	return RGBA(lhs.r / rhs.r, lhs.g / rhs.g, lhs.b / rhs.b, lhs.a / rhs.a);
}

// Vergleichsoperator
inline bool operator ==(const RGBA& lhs,
						const RGBA& rhs)
{
	return lhs.r == rhs.r && lhs.g == rhs.g && lhs.b == rhs.b && lhs.a == rhs.a;
}

// Kleiner-Operator (lexikografisch)
inline bool operator <(const RGBA& lhs,
					   const RGBA& rhs)
{
	if(lhs.r != rhs.r) return lhs.r < rhs.r;
	if(lhs.g != rhs.g) return lhs.g < rhs.g;
	if(lhs.b != rhs.b) return lhs.b < rhs.b;
	return lhs.a < rhs.a;
}

// Stream-Ausgabe einer Farbe
inline std::ostream& operator <<(std::ostream& lhs,
								 const RGBA& rhs)
{
	return lhs << '(' << rhs.r << ", " << rhs.g << ", " << rhs.b << ", " << rhs.a << ')';
}

// Stream-Eingabe einer Farbe,
// im Format (r, g, b, a) oder (r, g, b)
inline std::istream& operator >>(std::istream& lhs,
								 RGBA& rhs)
{
	// ffnende Klammer lesen
	char c = 0;
	lhs >> c;
	if(c != '(') lhs.clear(std::ios::badbit);

	// Farbwerte lesen, durch Kommata getrennt
	RGBA result;
	for(int i = 0; i < 4; ++i)
	{
		if(i)
		{
			lhs >> c;
			if(i == 3 && c == ')')
			{
				// Alpha standardmig 1
				result.a = 1.0f;
				lhs.putback(c);
				break;
			}
			else if(c != ',') lhs.clear(std::ios::badbit);
		}

		lhs >> result[i];
	}

	// schlieende Klammer lesen
	lhs >> c;
	if(c != ')') lhs.clear(std::ios::badbit);

	if(lhs)
	{
		// Ergebnis kopieren
		rhs = result;
	}

	return lhs;
}

inline const RGBA RGBA::fromRGBA32(Uint32 rgba32)
{
	// Bits zerlegen
	const int r8 = rgba32 & 0xFF;
	const int g8 = (rgba32 >> 8) & 0xFF;
	const int b8 = (rgba32 >> 16) & 0xFF;
	const int a8 = (rgba32 >> 24) & 0xFF;

	// auf [0, 1] abbilden
	return RGBA(static_cast<float>(r8) / 255.0f,
				static_cast<float>(g8) / 255.0f,
				static_cast<float>(b8) / 255.0f,
				static_cast<float>(a8) / 255.0f);
}

#endif
