/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "level.h"
#include "playerobject.h"
#include "simpleobject.h"

SimpleObject::SimpleObject(Layer& layer,
						   const TiXmlElement* p_xml,
						   const ParameterMap& gameParams)
	: Object(layer, p_xml, gameParams)
	, timeToLive(0.0f)
{
	// Typ lesen
	type = p_xml->Attribute("type");

	if(type == "Box")
	{
		visibilityRadius = 1.0f;
		level.addResourceToCache(sprites = app.getTextureManager().getResourceHandle("sprites/box.png"));
		b2BodyDef bodyDef;
		b2Body& body = createBody(bodyDef);
		b2PolygonDef polygonDef;
		polygonDef.SetAsBox(0.5f, 0.5f);
		polygonDef.density = 0.1f;
		polygonDef.friction = 0.5f;
		body.CreateShape(&polygonDef);
		body.SetMassFromShapes();
	}
	else if(type == "Nut")
	{
		visibilityRadius = 1.0f;
		level.addResourceToCache(sprites = app.getTextureManager().getResourceHandle("sprites/nut.png"));
		b2BodyDef bodyDef;
		b2Body& body = createBody(bodyDef);
		b2CircleDef circleDef;
		circleDef.radius = 0.5f;
		circleDef.isSensor = true;
		body.CreateShape(&circleDef);

		if(getGameMode() != GM_STORY) kill();
	}
	else if(type == "Sprite")
	{
		std::string file(p_xml->Attribute("file"));
		level.addResourceToCache(sprites = app.getTextureManager().getResourceHandle("sprites/" + file));
		const Vec2f size(fromString<Vec2f>(p_xml->Attribute("size")));
		params.set<Vec2f>("size", size);
		const char* color = p_xml->Attribute("color");
		if(color) params.set<RGBA>("color", fromString<RGBA>(color));
		else params.set<RGBA>("color", RGBA::WHITE);
		visibilityRadius = std::max(size.x, size.y);
	}
	else if(type == "Medkit")
	{
		visibilityRadius = 1.0f;
		level.addResourceToCache(sprites = app.getTextureManager().getResourceHandle("sprites/medkit.png"));
		b2BodyDef bodyDef;
		b2Body& body = createBody(bodyDef);
		b2PolygonDef polygonDef;
		polygonDef.SetAsBox(0.5f, 0.5f);
		polygonDef.isSensor = true;
		body.CreateShape(&polygonDef);
		params.set<int>("amount", fromString<int>(p_xml->Attribute("amount")));
	}
	else if(type == "Accelerator")
	{
		visibilityRadius = 1.0f;
		level.addResourceToCache(sprites = app.getTextureManager().getResourceHandle("sprites/accelerator.png"));
		b2BodyDef bodyDef;
		b2Body& body = createBody(bodyDef);
		b2PolygonDef polygonDef;
		polygonDef.SetAsBox(0.5f, 0.5f);
		polygonDef.isSensor = true;
		body.CreateShape(&polygonDef);

		float speed = 0.0f;
		p_xml->QueryFloatAttribute("speed", &speed);
		params.set<float>("speed", speed);

		SoundHandle sound = app.getSoundManager().getResourceHandle("sounds/accelerator.ogg");
		level.addResourceToCache(sound);
		params.set<SoundHandle>("sound", sound);
		params.set<SoundInstanceHandle>("soundInst", SoundInstanceHandle(0));
	}
	else if(type == "LayerSwitch")
	{
		visibilityRadius = 1.0f;
		level.addResourceToCache(sprites = app.getTextureManager().getResourceHandle("sprites/spiral.png"));
		b2BodyDef bodyDef;
		b2Body& body = createBody(bodyDef);
		b2CircleDef circleDef;
		circleDef.radius = 0.5f;
		circleDef.isSensor = true;
		body.CreateShape(&circleDef);
		params.set<std::string>("targetLayer", p_xml->Attribute("targetLayer"));
	}
	else
	{
		// ungltiger Typ
		assert(false);
	}

	// Initialisierung fertigstellen
	finishInit(p_xml);
}

void SimpleObject::update()
{
	if(timeToLive > 0.0f)
	{
		timeToLive -= app.getUpdateStepSize();
		if(timeToLive <= 0.0f)
		{
			// "Tod"
			kill();
		}
	}
}

void SimpleObject::render()
{
	app.setBlending(true);
	app.setTexturing(true);
	sprites->bind();

	if(type == "Box") drawBodyQuad(0, Vec2i(0, 0), Vec2i(128, 128), Vec2f(1.1f, 1.1f), false, false, RGBA(1.0f, 1.0f, 1.0f, timeToLive == 0.0f ? 1.0f : timeToLive));
	else if(type == "Nut") drawBodyQuad(0, Vec2i(0, 0), Vec2i(64, 64), Vec2f(0.75f, 0.75f));
	else if(type == "Sprite")
	{
		const Vec2f size = params.get<Vec2f>("size");
		const Vec2i txsize = sprites->getSize();
		const float ang = -deg2Rad(angle);
		const float sf = sinf(ang);
		const float cf = cosf(ang);
		float matrix[16] = { cf, -sf, 0.0f, 0.0f,
							 sf,  cf, 0.0f, 0.0f,
							 0.0f, 0.0f, 1.0f, 0.0f,
							 position.x, position.y, 0.0f, 1.0f };
		glPushMatrix();
		glMultMatrixf(matrix);
		RGBA color = params.get<RGBA>("color");
		glColor4fv(color);
		glBegin(GL_QUADS);
			glTexCoord2i(0, 0);
			glVertex2f(size.x * -0.5f, size.y * -0.5f);
			glTexCoord2i(txsize.x, 0);
			glVertex2f(size.x * 0.5f, size.y * -0.5f);
			glTexCoord2i(txsize.x, txsize.y);
			glVertex2f(size.x * 0.5f, size.y * 0.5f);
			glTexCoord2i(0, txsize.y);
			glVertex2f(size.x * -0.5f, size.y * 0.5f);
		glEnd();
		glPopMatrix();
	}
	else if(type == "Medkit") drawBodyQuad(0, Vec2i(0, 0), Vec2i(128, 128), Vec2f(2.0f, 2.0f));
	else if(type == "Accelerator") drawBodyQuad(0, Vec2i(0, 0), Vec2i(64, 64), Vec2f(1.0f, 1.0f));
	else if(type == "LayerSwitch") drawBodyQuad(0, Vec2i(0, 0), Vec2i(256, 256), Vec2f(1.0f, 1.0f));
}

void SimpleObject::onCollision(const Contact& contact)
{
	if(PlayerObject* p_player = dynamic_cast<PlayerObject*>(contact.p_otherObject))
	{
		if(type == "Nut")
		{
			p_player->addPoints(50);
			kill();
		}
		else if(type == "Box")
		{
			// 2 Sekunden bis zum "Tod"
			if(timeToLive == 0.0f) timeToLive = 2.0f;
		}
		else if(type == "Medkit" && p_player->getHealth() < 100)
		{
			p_player->reduceHealth(-params.get<int>("amount"));
			kill();
		}
		else if(type == "Accelerator")
		{
			p_player->addVelocity(getXAxis() * params.get<float>("speed"));

			// Sound abspielen
			SoundInstanceHandle soundInst = params.get<SoundInstanceHandle>("soundInst");
			if(!soundInst || !soundInst->isValid())
			{
				soundInst = params.get<SoundHandle>("sound")->createInstance();
				params.set<SoundInstanceHandle>("soundInst", soundInst);
			}
		}
		else if(type == "LayerSwitch")
		{
			p_player->setLayerDelayed(params.get<std::string>("targetLayer"));
		}
	}
}
