/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "sound.h"
#include "soundinstance.h"

void SoundManager::update()
{
	// alle Sounds aktualisieren
	for(ResourceMapType::iterator it = resources.begin();
		it != resources.end();
		++it)
	{
		it->second->update();
	}
}

Sound::Sound(SoundManager& manager,
			 const std::string& filename,
			 const ParameterMap& options)
	: Resource(manager, filename, options)
	, soundBufferID(0)
	, format(0)
	, frequency(0)
	, duration(0.0)
{
	// Datei zum binren Lesen ffnen
	FILE* p_file = fopen(filename.c_str(), "rb");
	assert(p_file);

	// Vorbis-Datei ffnen
	OggVorbis_File vorbisFile;
	int r = ov_open(p_file, &vorbisFile, 0, 0);
	assert(!r);

	// Informationen holen
	vorbis_info* p_info = ov_info(&vorbisFile, -1);
	assert(p_info);

	// Mono oder Stereo?
	format = p_info->channels == 1 ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16;

	// Frequenz holen
	frequency = p_info->rate;

	// Lnge holen
	duration = ov_time_total(&vorbisFile, -1);

	// Streamen oder an einem Stck laden?
	if(!getStreaming())
	{
		// den kompletten Sound dekodieren (jeweils in Stcken von 1024 Bytes)
		std::vector<char> bufferData;
		char chunk[1024];
		int bitStream = 0;
		while(true)
		{
			const long numBytesRead = ov_read(&vorbisFile, chunk, sizeof(chunk), 0, 2, 1, &bitStream);
			assert(numBytesRead >= 0);
			if(!numBytesRead) break;
			bufferData.insert(bufferData.end(), chunk, chunk + numBytesRead);
		}

		// Puffer sollte nicht leer sein ...
		assert(!bufferData.empty());

		// OpenAL-Buffer erzeugen
		alGenBuffers(1, &soundBufferID);
		assert(soundBufferID);

		// Sample-Daten in den Sound-Buffer schreiben
		alBufferData(soundBufferID, format, &bufferData.front(), static_cast<ALsizei>(bufferData.size()), frequency);
	}

	// aufrumen
	ov_clear(&vorbisFile);
}

Sound::~Sound()
{
	if(soundBufferID)
	{
		// OpenAL-Buffer zerstren
		alDeleteBuffers(1, &soundBufferID);
	}
}

SoundInstanceHandle Sound::createInstance(bool play,
										  bool loop,
										  float volume,
										  float pitch)
{
	// neue Instanz erzeugen und eintragen
	SoundInstanceHandle instance(new SoundInstance(*this, play, loop, volume, pitch));
	instances.push_back(instance);
	return instance;
}

void Sound::update()
{
	// alle Instanzen durchgehen und diejenigen entfernen,
	// die aufgerumt werden knnen
	for(SoundInstanceListType::iterator it = instances.begin();
		it != instances.end();)
	{
		if(!(*it)->update())
		{
			// entfernen (Referenzzhler wird verringert)
			it = instances.erase(it);
		}
		else
		{
			// weiter
			++it;
		}
	}
}

ALuint Sound::getSoundBufferID() const
{
	return soundBufferID;
}

bool Sound::getStreaming() const
{
	return options.get<bool>("streaming");
}

ALenum Sound::getFormat() const
{
	return format;
}

ALsizei Sound::getFrequency() const
{
	return frequency;
}

double Sound::getDuration() const
{
	return duration;
}

const ParameterMap Sound::getDefaultOptions()
{
	ParameterMap options;

	// Standard: Sound komplett in einen Puffer dekodieren
	options.set<bool>("streaming", false);

	return options;
}
