/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "sound.h"
#include "soundinstance.h"

SoundInstance::SoundInstance(Sound& sound,
							 bool play,
							 bool loop,
							 float volume,
							 float pitch)
	: sound(sound)
	, sourceID(0)
	, streaming(sound.getStreaming())
	, loop(false)
	, endOfStreamReached(false)
	, p_streamingThread(0)
	, p_streamingThreadExit(0)
	, fading(false)
	, fadeStartVolume(0.0f)
	, fadeStartPitch(0.0f)
	, fadeTargetVolume(0.0f)
	, fadeTargetPitch(0.0f)
	, fadeDuration(0.0f)
	, fadeTime(0.0f)
	, fadeCurve(0.0f)
	, fadeEndAction(FAE_NOTHING)
{
	// Soundquelle erzeugen
	alGenSources(1, &sourceID);
	assert(sourceID);

	if(streaming)
	{
		// Vorbis-Datei ffnen
		FILE* p_file = fopen(sound.getFilename().c_str(), "rb");
		assert(p_file);
		int r = ov_open(p_file, &vorbisFile, 0, 0);
		assert(!r);

		// Mutex erzeugen
		p_streamingThreadExit = SDL_CreateSemaphore(0);

		// Thread zum Streamen erzeugen
		p_streamingThread = SDL_CreateThread(streamingThreadProc, this);
	}
	else
	{
		// Dieser Sound wurde bereits vollstndig geladen.
		// Er kann sofort abgespielt werden, ohne Streaming.
		alSourcei(sourceID, AL_BUFFER, sound.getSoundBufferID());
	}

	setLoop(loop);
	setVolume(volume);
	setPitch(pitch);
	if(play) this->play();
}

SoundInstance::~SoundInstance()
{
	if(isValid()) cleanUp();
}

void SoundInstance::play()
{
	assert(isValid());

	if(streaming)
	{
		// wieder von vorne anfangen, Puffer lschen und einen neuen Puffer dekodieren
		ov_time_seek(&vorbisFile, 0.0);
		endOfStreamReached = false;
		unqueueAndDeleteAllBuffers();
		streamBuffer();
	}

	// Soundquelle zurcksetzen und abspielen
	alSourceRewind(sourceID);
	alSourcePlay(sourceID);
}

void SoundInstance::stop()
{
	assert(isValid());

	// stoppen und sofort aufrumen
	alSourceStop(sourceID);
	cleanUp();
}

void SoundInstance::pause()
{
	assert(isValid());
	alSourcePause(sourceID);
}

void SoundInstance::resume()
{
	assert(isValid());
	alSourcePlay(sourceID);
}

void SoundInstance::setLoop(bool enable)
{
	assert(isValid());

	if(!streaming)
	{
		// Looping nur fr nicht gestreamte Sounds
		alSourcei(sourceID, AL_LOOPING, enable ? AL_TRUE : AL_FALSE);
	}

	loop = enable;
}

void SoundInstance::setVolume(float volume)
{
	assert(isValid());
	alSourcef(sourceID, AL_GAIN, volume);
}

void SoundInstance::setPitch(float pitch)
{
	assert(isValid());
	alSourcef(sourceID, AL_PITCH, pitch);
}

void SoundInstance::fade(float targetVolume,
						 float targetPitch,
						 float duration,
						 float curve,
						 FadeEndAction endAction)
{
	assert(isValid());

	fadeStartVolume = getVolume();
	fadeStartPitch = getPitch();
	fadeTargetVolume = targetVolume;
	fadeTargetPitch = targetPitch;
	fadeDuration = duration;
	fadeTime = 0.0f;
	fadeCurve = curve;
	fadeEndAction = endAction;
	fading = true;
}

void SoundInstance::stopFade()
{
	assert(isValid());
	fading = false;
}

bool SoundInstance::isPlaying() const
{
	assert(isValid());

	ALint state = 0;
	alGetSourcei(sourceID, AL_SOURCE_STATE, &state);
	return state == AL_PLAYING;
}

bool SoundInstance::isPaused() const
{
	assert(isValid());

	ALint state = 0;
	alGetSourcei(sourceID, AL_SOURCE_STATE, &state);
	return state == AL_PAUSED;
}

bool SoundInstance::getLoop() const
{
	assert(isValid());
	return loop;
}

float SoundInstance::getVolume() const
{
	assert(isValid());

	ALfloat volume = 0.0f;
	alGetSourcef(sourceID, AL_GAIN, &volume);
	return volume;
}

float SoundInstance::getPitch() const
{
	assert(isValid());

	ALfloat pitch = 0.0f;
	alGetSourcef(sourceID, AL_PITCH, &pitch);
	return pitch;
}

bool SoundInstance::isFading() const
{
	assert(isValid());
	return fading;
}

bool SoundInstance::isValid() const
{
	return sourceID != 0;
}

bool SoundInstance::update()
{
	if(!isValid()) return false;

	ALint state = 0;
	alGetSourcei(sourceID, AL_SOURCE_STATE, &state);
	if(state == AL_STOPPED)
	{
		if(streaming && !endOfStreamReached)
		{
			// Die Wiedergabe wurde beendet, weil der Thread
			// wahrscheinlich nicht schnell genug dekodieren konnte.
			// Einen Puffer dekodieren und die Wiedergabe wieder starten.
			unqueueAndDeleteAllBuffers();
			streamBuffer();
			alSourcePlay(sourceID);
		}
		else
		{
			// Soundinstanz kann gelscht werden
			cleanUp();
			return false;
		}
	}

	// Fading verarbeiten
	if(fading && isPlaying())
	{
		// Ziel erreicht?
		fadeTime += sound.getManager().getApplication().getUpdateStepSize();
		if(fadeTime >= fadeDuration)
		{
			// Zielwerte setzen und Fading beenden
			setVolume(fadeTargetVolume);
			setPitch(fadeTargetPitch);
			fading = false;

			// Was soll jetzt passieren?
			switch(fadeEndAction)
			{
			case FAE_NOTHING:
				// Nichts!
				break;
			case FAE_PAUSE:
				// pausieren
				pause();
				break;
			case FAE_STOP:
				// stoppen, woraufhin die Instanz entfernt wird
				stop();
				return false;
			}
		}
		else
		{
			// Fading
			const float t = pow(fadeTime / fadeDuration, fadeCurve);
			const float newVolume = fadeTargetVolume * t + fadeStartVolume * (1.0f - t);
			const float newPitch = fadeTargetPitch * t + fadeStartPitch * (1.0f - t);
			setVolume(newVolume);
			setPitch(newPitch);
		}
	}

	// Soundinstanz ist noch nicht fertig
	return true;
}

void SoundInstance::streamBuffer()
{
	// mindestens 0.25 Sekunden dekodieren
	double timeLeft = 0.25;
	std::vector<char> bufferData;
	char chunk[1024];
	int bitStream = 0;
	while(timeLeft > 0.0)
	{
		// Daten lesen
		const double t0 = ov_time_tell(&vorbisFile);
		const long numBytesRead = ov_read(&vorbisFile, chunk, sizeof(chunk), 0, 2, 1, &bitStream);
		assert(numBytesRead >= 0);
		const double t1 = ov_time_tell(&vorbisFile);
		timeLeft -= t1 - t0;

		if(!numBytesRead)
		{
			// Ende der Datei erreicht!
			// Wenn Looping aktiviert ist, dann wieder von vorne anfangen.
			if(loop) ov_time_seek(&vorbisFile, 0.0);
			else
			{
				endOfStreamReached = true;
				break;
			}
		}

		// Daten einfgen
		bufferData.insert(bufferData.end(), chunk, chunk + numBytesRead);
	}

	if(!bufferData.empty())
	{
		// Puffer erzeugen und Daten hinein schreiben
		ALuint soundBufferID = 0;
		alGenBuffers(1, &soundBufferID);
		assert(soundBufferID);
		alBufferData(soundBufferID, sound.getFormat(), &bufferData.front(), static_cast<ALsizei>(bufferData.size()), sound.getFrequency());

		// Puffer an die Quelle anhngen
		alSourceQueueBuffers(sourceID, 1, &soundBufferID);
	}
}

void SoundInstance::unqueueAndDeleteAllBuffers()
{
	// alle Puffer aus der Warteschlange nehmen und lschen
	alSourceStop(sourceID);
	ALint numBuffersPlayed = 0;
	alGetSourcei(sourceID, AL_BUFFERS_PROCESSED, &numBuffersPlayed);
	if(numBuffersPlayed)
	{
		ALuint* p_buffers = new ALuint[numBuffersPlayed];
		alSourceUnqueueBuffers(sourceID, numBuffersPlayed, p_buffers);
		alDeleteBuffers(numBuffersPlayed, p_buffers);
		delete[] p_buffers;
	}
}

void SoundInstance::cleanUp()
{
	assert(isValid());

	if(streaming)
	{
		// Thread soll beendet werden
		SDL_SemPost(p_streamingThreadExit);
		SDL_WaitThread(p_streamingThread, 0);

		// Semaphore lschen
		SDL_DestroySemaphore(p_streamingThreadExit);

		// Puffer lschen
		unqueueAndDeleteAllBuffers();

		// Vorbis-Datei schlieen
		ov_clear(&vorbisFile);
	}

	alDeleteSources(1, &sourceID);
	sourceID = 0;
}

int SoundInstance::streamingThreadProc(void* p_arg)
{
	SoundInstance& me = *(reinterpret_cast<SoundInstance*>(p_arg));

	while(true)
	{
		// maximal 125 ms auf das Signal zum Beenden des Threads warten,
		// dann prfen, ob mehr Daten bentigt werden
		if(!SDL_SemWaitTimeout(me.p_streamingThreadExit, 125))
		{
			// Der Thread soll sich beenden.
			break;
		}

#ifdef __MACOSX__
		// Semaphore freigeben
		SDL_SemPost(me.p_streamingThreadExit);
#endif

		// abgespielte Puffer aus der Warteschlange nehmen und lschen
		ALint numBuffersPlayed = 0;
		alGetSourcei(me.sourceID, AL_BUFFERS_PROCESSED, &numBuffersPlayed);
		if(numBuffersPlayed)
		{
			ALuint* p_buffers = new ALuint[numBuffersPlayed];
			alSourceUnqueueBuffers(me.sourceID, numBuffersPlayed, p_buffers);
			alDeleteBuffers(numBuffersPlayed, p_buffers);
			delete[] p_buffers;
		}

		if(me.isPlaying() || me.isPaused())
		{
			// Anzahl der Puffer in der Warteschlange holen
			ALint numBuffersQueued = 0;
			alGetSourcei(me.sourceID, AL_BUFFERS_QUEUED, &numBuffersQueued);

			// Neue Puffer bentigt? Wenn ja, dann dekodieren und anhngen.
			while(numBuffersQueued++ < 4) me.streamBuffer();
		}
	}

	return 0;
}
