/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#ifndef SOUNDINSTANCE_H_INCLUDED
#define SOUNDINSTANCE_H_INCLUDED

class Sound;
class SoundInstance;

// Handle fr Soundinstanzen
typedef Handle<SoundInstance> SoundInstanceHandle;

// Klasse fr Instanzen eines Sounds
class SoundInstance
	: NonCopyable
	, public ReferenceCounted
{
public:
	// Optionen fr Aktionen nach Ende eines Fades
	enum FadeEndAction
	{
		FAE_NOTHING,	// nichts tun
		FAE_PAUSE,		// Sound pausieren
		FAE_STOP		// Sound stoppen
	};

	// Konstruktor
	SoundInstance(Sound& sound, bool play, bool loop, float volume, float pitch);

	// Destruktor
	~SoundInstance();

	// spielt die Soundinstanz von Beginn an ab
	void play();

	// pausiert die Soundinstanz
	void pause();

	// setzt die Soundinstanz nach einer Pause fort
	void resume();

	// stoppt die Soundinstanz, woraufhin sie vom Manager entfernt wird
	// (nach diesem Aufruf ist die Soundinstanz ungltig)
	void stop();

	// schaltet Looping ein oder aus
	void setLoop(bool enable);

	// setzt die Lautstrke (1.0 ist die normale Lautstrke)
	void setVolume(float volume);

	// setzt die Tonhhe (1.0 ist die normale Frequenz)
	void setPitch(float pitch);

	// passt die Lautstrke und die Tonhhe der Soundinstanz innerhalb der angegebenen Zeit
	// an den gegebenen Zielwert an und stoppt danach gegebenenfalls die Soundinstanz
	void fade(float targetVolume, float targetPitch, float duration, float curve = 1.0f, FadeEndAction endAction = FAE_NOTHING);

	// stoppt das Fading
	void stopFade();

	// prft, ob die Soundinstanz gerade abgespielt wird
	bool isPlaying() const;

	// prft, ob die Soundinstanz gerade pausiert ist
	bool isPaused() const;

	// prft, ob Looping aktiviert ist
	bool getLoop() const;

	// liefert die Lautstrke
	float getVolume() const;

	// liefert die Tonhhe
	float getPitch() const;

	// prft, ob die Soundinstanz gerade gefadet wird
	bool isFading() const;

	// prft, ob die Soundinstanz noch gltig ist
	bool isValid() const;

	// aktualisiert die Soundinstanz und liefert false, wenn sie fertig ist
	// und gelscht werden kann (vom jeweiligen Sound)
	bool update();

private:
	// einen Soundpuffer dekodieren und an die Soundquelle anhngen
	void streamBuffer();

	// nimmt alle Puffer aus der Warteschlange und zerstrt sie
	void unqueueAndDeleteAllBuffers();

	// aufrumen
	void cleanUp();

	// Funktion, die im Streaming-Thread ausgefhrt wird
	static int streamingThreadProc(void* p_arg);

	Sound& sound;					// der Sound
	ALuint sourceID;				// ID der Soundquelle
	bool streaming;					// Wird der Sound gestreamt?
	bool loop;						// Looping?
	bool endOfStreamReached;		// Wurde das Ende des Streams erreicht?
	OggVorbis_File vorbisFile;		// Vorbis-Datei (nur bei Sounds mit Streaming)
	SDL_Thread* p_streamingThread;	// Thread zum Streamen
	SDL_sem* p_streamingThreadExit;	// Bedingung zum Beenden des Streaming-Threads
	bool fading;					// Fading gerade aktiv?
	float fadeStartVolume;			// Startwert der Lautstrke
	float fadeStartPitch;			// Startwert der Tonhhe
	float fadeTargetVolume;			// Zielwert der Lautstrke
	float fadeTargetPitch;			// Zielwert der Tonhhe
	float fadeDuration;				// Dauer des Fadings
	float fadeTime;					// Fortschritt des Fadings
	float fadeCurve;				// Exponent der Fading-Kurve
	FadeEndAction fadeEndAction;	// Was tun, wenn Fade beendet ist?
};

#endif
