/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "texture.h"

Texture::Texture(TextureManager& manager,
				 const std::string& filename,
				 const ParameterMap& options)
	: Resource(manager, filename, options)
	, textureID(0)
	, size(0, 0)
{
	// Bild laden
	SDL_Surface* p_surface = IMG_Load(filename.c_str());
	assert(p_surface);

	// Gre eintragen
	size = Vec2i(p_surface->w, p_surface->h);

	// Surface mit fest definiertem Farbformat erzeugen
	SDL_Surface* p_rgbaSurface = SDL_CreateRGBSurface(SDL_SWSURFACE, p_surface->w, p_surface->h, 32, 0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000);
	assert(p_rgbaSurface);

	// Bild in die neue Surface kopieren (Farben werden konvertiert)
	SDL_SetAlpha(p_surface, 0, 0);
	SDL_SetAlpha(p_rgbaSurface, 0, 0);
	SDL_BlitSurface(p_surface, 0, p_rgbaSurface, 0);

	// Das alte Bild brauchen wir nicht mehr.
	SDL_FreeSurface(p_surface);

	// OpenGL-Textur erzeugen
	glGenTextures(1, &textureID);
	assert(textureID);

	// alle OpenGL-Zustnde speichern
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	// Textur aktivieren
	glBindTexture(GL_TEXTURE_2D, textureID);

	// Sollen MIP-Maps generiert werden?
	const bool mipMapping = options.get<bool>("mipMapping");

	// Filtertyp holen
	const DownscaleFilter downscaleFilter = options.get<DownscaleFilter>("downscaleFilter");

	// Texturfilter setzen
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mipMapping ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	int mipLevel = 0;
	while(p_rgbaSurface)
	{
		// Surface sperren und Parameter einstellen, um auf die Pixel zugreifen zu knnen
		SDL_LockSurface(p_rgbaSurface);
		glPixelStorei(GL_UNPACK_ROW_LENGTH, p_rgbaSurface->pitch / p_rgbaSurface->format->BytesPerPixel);

		if(mipMapping && p_rgbaSurface->w != p_rgbaSurface->h)
		{
			// das muss GLU fr uns erledigen ...
			gluBuild2DMipmaps(GL_TEXTURE_2D, 4, p_rgbaSurface->w, p_rgbaSurface->h, GL_RGBA, GL_UNSIGNED_BYTE, p_rgbaSurface->pixels);
			SDL_FreeSurface(p_rgbaSurface);
			break;
		}

		// Pixel in die OpenGL-Textur kopieren
		glTexImage2D(GL_TEXTURE_2D, mipLevel, GL_RGBA, p_rgbaSurface->w, p_rgbaSurface->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_rgbaSurface->pixels);

		// Surface entsperren
		SDL_UnlockSurface(p_rgbaSurface);

		if(mipMapping && p_rgbaSurface->w > 1 && p_rgbaSurface->h > 1)
		{
			// nchsten MIP-Map-Level vorbereiten
			SDL_Surface* p_nextLevel = downscaleSurface2x(p_rgbaSurface, downscaleFilter);
			SDL_FreeSurface(p_rgbaSurface);
			p_rgbaSurface = p_nextLevel;
			++mipLevel;
		}
		else
		{
			// Jetzt ist Schluss!
			SDL_FreeSurface(p_rgbaSurface);
			p_rgbaSurface = 0;
		}
	}

	// Texturmatrix so einstellen, dass die Texturkoordinaten Pixel sind
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glLoadIdentity();
	glScalef(1.0f / size.x, 1.0f / size.y, 1.0f);
	glGetFloatv(GL_TEXTURE_MATRIX, textureMatrix);
	glPopMatrix();

	// alten Zustand wiederherstellen
	glPopAttrib();
}

Texture::~Texture()
{
	// Textur freigeben
	glDeleteTextures(1, &textureID);
}

void Texture::bind() const
{
	// Textur binden
	glBindTexture(GL_TEXTURE_2D, textureID);

	// Texturmatrix setzen
	GLint matrixMode = 0;
	glGetIntegerv(GL_MATRIX_MODE, &matrixMode);
	glMatrixMode(GL_TEXTURE);
	glLoadMatrixf(textureMatrix);
	glMatrixMode(matrixMode);
}

const Vec2i& Texture::getSize() const
{
	return size;
}

GLuint Texture::getTextureID() const
{
	return textureID;
}

const ParameterMap Texture::getDefaultOptions()
{
	ParameterMap options;

	// Standard: MIP-Mapping aktivieren
	// (MIP-Maps werden beim Laden generiert)
	options.set<bool>("mipMapping", true);

	// Standard: normaler Box-Filter zum Herunterskalieren der Textur
	options.set<DownscaleFilter>("downscaleFilter", DF_BOX);

	return options;
}
