/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "layer.h"
#include "tile.h"

Tile::Tile(const TiXmlElement* p_xml,
		   const Vec2i& defaultSize,
		   float defaultShapeRestitution,
		   float defaultShapeFriction)
	: id(0)
	, position(0, 0)
	, size(defaultSize)
	, isSimple1x1Box(false)
	, visible(true)
{
	// ID lesen
	p_xml->QueryIntAttribute("id", &id);

	// Position lesen
	position = fromString<Vec2i>(p_xml->Attribute("position"));

	// Gre lesen
	if(const char* p_size = p_xml->Attribute("size")) size = fromString<Vec2i>(p_size);

	// Sichtbarkeit lesen
	int visible = 1;
	p_xml->QueryIntAttribute("visible", &visible);
	this->visible = visible != 0;

	// Shapes lesen
	if(const TiXmlElement* p_shapesXML = p_xml->FirstChildElement("Shapes"))
	{
		for(const TiXmlElement* p_shapeXML = p_shapesXML->FirstChildElement();
			p_shapeXML;
			p_shapeXML = p_shapeXML->NextSiblingElement())
		{
			b2ShapeDef* p_shapeDef = 0;

			// Typ des Shapes?
			const std::string shapeType(p_shapeXML->Value());
			if(shapeType == "PolygonShape")
			{
				b2PolygonDef* p_polygonDef = new b2PolygonDef;
				std::istringstream stream(p_shapeXML->Attribute("vertices"));
				while(stream)
				{
					// Vektor lesen
					Vec2f vertex;
					stream >> vertex;
					p_polygonDef->vertices[p_polygonDef->vertexCount++] = vertex;

					// Komma lesen
					char c;
					stream >> c;
				}

				p_shapeDef = p_polygonDef;
			}
			else if(shapeType == "Simple1x1BoxShape")
			{
				b2PolygonDef* p_polygonDef = new b2PolygonDef;
				p_polygonDef->vertices[p_polygonDef->vertexCount++] = b2Vec2(0.0f, 0.0f);
				p_polygonDef->vertices[p_polygonDef->vertexCount++] = b2Vec2(1.0f, 0.0f);
				p_polygonDef->vertices[p_polygonDef->vertexCount++] = b2Vec2(1.0f, 1.0f);
				p_polygonDef->vertices[p_polygonDef->vertexCount++] = b2Vec2(0.0f, 1.0f);
				p_shapeDef = p_polygonDef;
				isSimple1x1Box = true;
			}
			else
			{
				// ungltiger Typ
				assert(false);
			}

			assert(p_shapeDef);

			// Elastizitt lesen
			p_shapeDef->restitution = defaultShapeRestitution;
			p_shapeXML->QueryFloatAttribute("restitution", &p_shapeDef->restitution);

			// Reibung lesen
			p_shapeDef->friction = defaultShapeFriction;
			p_shapeXML->QueryFloatAttribute("friction", &p_shapeDef->friction);

			// Shape hinzufgen
			shapeDefs.push_back(p_shapeDef);
		}
	}

	if(shapeDefs.size() > 1) isSimple1x1Box = false;
}

Tile::~Tile()
{
	// Shape-Definitionen lschen
	for(ShapeDefListType::const_iterator it = shapeDefs.begin();
		it != shapeDefs.end();
		++it)
	{
		delete *it;
	}
}

void Tile::draw(const Vec2f& position,
				const RGBA& color) const
{
	glColor4fv(color);
	glTexCoord2i(this->position.x, this->position.y);
	glVertex2f(position.x, position.y);
	glTexCoord2i(this->position.x + size.x, this->position.y);
	glVertex2f(position.x + 1.0f, position.y);
	glTexCoord2i(this->position.x + size.x, this->position.y + size.y);
	glVertex2f(position.x + 1.0f, position.y + 1.0f);
	glTexCoord2i(this->position.x, this->position.y + size.y);
	glVertex2f(position.x, position.y + 1.0f);
}

void Tile::createShapes(b2Body& body,
						Layer& layer,
						const Vec2f& position,
						const Vec2i& blockSize) const
{
	for(ShapeDefListType::const_iterator it = shapeDefs.begin();
		it != shapeDefs.end();
		++it)
	{
		// Shape-Definition transformieren
		b2ShapeDef* p_transformedShapeDef = transformShapeDef(**it, position, blockSize);

		// Kollisions-Bits setzen
		p_transformedShapeDef->filter.categoryBits = p_transformedShapeDef->filter.maskBits = 1 << layer.getIndex();

		// Shape erzeugen
		body.CreateShape(p_transformedShapeDef);

		// wieder lschen
		delete p_transformedShapeDef;
	}
}

int Tile::getID() const
{
	return id;
}

bool Tile::getIsSimple1x1Box() const
{
	return isSimple1x1Box;
}

bool Tile::isVisible() const
{
	return visible;
}
