/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "level.h"
#include "tile.h"
#include "tilemaplayer.h"
#include "tileset.h"

TileMapLayer::TileMapLayer(Level& level,
						   int index,
						   const TiXmlElement* p_xml,
						   const ParameterMap& gameParams)
	: Layer(level, index, p_xml, gameParams)
	, p_tileSet(0)
	, size(0, 0)
	, p_tiles(0)
	, enablePhysics(false)
{
	// Dateiname des Tile-Sets lesen
	const std::string tileSetFilename(p_xml->Attribute("tileSetFilename"));

	// Tile-Set laden
	TiXmlDocument doc;
	doc.LoadFile(app.getFileLocator().findFile("tilesets/" + tileSetFilename).c_str());
	p_tileSet = new TileSet(app, doc.FirstChildElement("TileSet"));

	// Physik lesen
	int enablePhysics = 0;
	p_xml->QueryIntAttribute("enablePhysics", &enablePhysics);
	this->enablePhysics = enablePhysics != 0;

	// die einzelnen Zeilen laden
	Vec2i cursor(0, 0);
	std::map<Vec2i, Uint32> tileMap;
	for(const TiXmlElement* p_row = p_xml->FirstChildElement("Row");
		p_row;
		p_row = p_row->NextSiblingElement("Row"), ++cursor.y)
	{
		const std::string data(p_row->GetText());
		std::istringstream stream(data);
		cursor.x = 0;

		// Zeile lesen
		while(true)
		{
			// ID lesen
			stream >> tileMap[cursor];
			if(stream)
			{
				// weitere Spalte
				++cursor.x;
				size.x = std::max(size.x, cursor.x);

				// Komma lesen
				char c = 0;
				stream >> c;
			}
			else
			{
				// Ende der Zeile
				++size.y;
				break;
			}
		}
	}

	// Tiles zeilenweise speichern
	p_tiles = new Uint32[size.x * size.y];
	for(int y = 0; y < size.y; ++y)
	{
		for(int x = 0; x < size.x; ++x)
		{
			// Tile suchen und eintragen, ansonsten 0 eintragen (leer)
			std::map<Vec2i, Uint32>::const_iterator it = tileMap.find(Vec2i(x, y));
			p_tiles[y * size.x + x] = it == tileMap.end() ? 0 : it->second;
		}
	}
}

TileMapLayer::~TileMapLayer()
{
	// Tiles lschen
	delete[] p_tiles;

	// Tile-Set lschen
	delete p_tileSet;
}

void TileMapLayer::update()
{
	updateObjects();
}

void TileMapLayer::render(Object* p_forceRenderingOnTop)
{
	// Blending und Transformation setzen
	setupBlending();
	glPushMatrix();
	setupMatrix();

	// herausfinden, welcher Teil der Tile-Map sichtbar ist,
	// und dann auch nur diesen Teil rendern
	double z;
	double coords[3];
	double modelViewMatrix[16];
	double projectionMatrix[16];
	GLint viewport[4];
	glGetDoublev(GL_MODELVIEW_MATRIX, modelViewMatrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projectionMatrix);
	glGetIntegerv(GL_VIEWPORT, viewport);
	gluProject(0.0, 0.0, 0.0, modelViewMatrix, projectionMatrix, viewport, coords, coords + 1, &z);
	gluUnProject(viewport[0], viewport[1] + viewport[3], z, modelViewMatrix, projectionMatrix, viewport, coords, coords + 1, coords + 2);
	Vec2i topLeft(Vec2d(coords[0], coords[1]));
	topLeft.x = std::max(topLeft.x, 0);
	topLeft.y = std::max(topLeft.y, 0);
	gluUnProject(viewport[0] + viewport[2], viewport[1], z, modelViewMatrix, projectionMatrix, viewport, coords, coords + 1, coords + 2);
	Vec2i bottomRight(Vec2d(coords[0], coords[1]));
	bottomRight.x = std::min(bottomRight.x, size.x - 1);
	bottomRight.y = std::min(bottomRight.y, size.y - 1);

	// Tiles rendern
	glPushMatrix();
	glTranslated(topLeft.x, topLeft.y, 0.0);
	p_tileSet->drawTiles(p_tiles + topLeft.y * size.x + topLeft.x, bottomRight - topLeft + Vec2i(1, 1), size.x, color);
	glPopMatrix();

	// Objekte rendern
	renderObjects(p_forceRenderingOnTop);

	glPopMatrix();
}

void TileMapLayer::createShapes(b2Body& body)
{
	if(!enablePhysics)
	{
		// nichts zu tun, weil keine Physik aktiviert ist
		return;
	}

	bool* p_tilesProcessed = new bool[size.x * size.y];
	for(int i = 0; i < size.x * size.y; ++i) p_tilesProcessed[i] = false;

	// alle Tiles durchgehen
	for(int y = 0; y < size.y; ++y)
	{
		for(int x = 0; x < size.x; ++x)
		{
			const Tile* p_tile = getTileAt(Vec2i(x, y));
			if(p_tile && !p_tilesProcessed[y * size.x + x])
			{
				Vec2i blockSize(1, 1);

				const Tile* p_tile = getTileAt(Vec2i(x, y));
				if(p_tile->getIsSimple1x1Box())
				{
					// versuchen, so viele dieser Tiles wie mglich zu einem Block zusammenzufassen
					int axes = 1 | 2;
					do extendBlock(Vec2i(x, y), blockSize, axes); while(axes);
				}

				// Shapes an der richtigen Stelle erzeugen lassen
				p_tile->createShapes(body, *this, Vec2i(x, y), blockSize);

				// Tiles dieses Blocks als verarbeitet markieren
				for(int x2 = x; x2 < x + blockSize.x; ++x2)
					for(int y2 = y; y2 < y + blockSize.y; ++y2)
						p_tilesProcessed[y2 * size.x + x2] = true;
			}
		}
	}

	delete[] p_tilesProcessed;
}

int TileMapLayer::getTileIDAt(const Vec2i& position) const
{
	if(position.x < 0 || position.y < 0 || position.x >= size.x || position.y >= size.y) return 0;
	else return p_tiles[position.y * size.x + position.x];
}

const Tile* TileMapLayer::getTileAt(const Vec2i& position) const
{
	if(position.x < 0 || position.y < 0 || position.x >= size.x || position.y >= size.y) return 0;
	else return p_tileSet->getTileByID(p_tiles[position.y * size.x + position.x]);
}

void TileMapLayer::extendBlock(const Vec2i& position,
							   Vec2i& blockSize,
							   int& axes) const
{
	if(axes & 1)
	{
		// versuchen, auf der x-Achse zu vergrern
		const int nx = position.x + blockSize.x;
		for(int y = position.y; y < position.y + blockSize.y; ++y)
		{
			const Tile* p_tile = getTileAt(Vec2i(nx, y));
			if(!p_tile || !p_tile->getIsSimple1x1Box())
			{
				axes &= ~1;
				break;
			}
		}

		if(axes & 1) ++blockSize.x;
	}

	if(axes & 2)
	{
		// versuchen, auf der y-Achse zu vergrern
		const int ny = position.y + blockSize.y;
		for(int x = position.x; x < position.x + blockSize.x; ++x)
		{
			const Tile* p_tile = getTileAt(Vec2i(x, ny));
			if(!p_tile || !p_tile->getIsSimple1x1Box())
			{
				axes &= ~2;
				break;
			}
		}

		if(axes & 2) ++blockSize.y;
	}
}
