/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"
#include "application.h"
#include "tile.h"
#include "tileset.h"

TileSet::TileSet(Application& app,
				 const TiXmlElement* p_xml)
	: app(app)
{
	// Dateiname des Bilds holen und das Bild laden
	const std::string imageFilename(p_xml->Attribute("imageFilename"));
	image = app.getTextureManager().getResourceHandle("tilesets/" + imageFilename);

	// Standardgre der Tiles lesen
	// (wird benutzt, wenn fr ein Tile keine Gre angegeben ist)
	Vec2i defaultTileSize(0, 0);
	if(const char* p_defaultTileSize = p_xml->Attribute("defaultTileSize")) defaultTileSize = fromString<Vec2i>(p_defaultTileSize);

	// Standardelastizitt lesen
	float defaultShapeRestitution = 0.0f;
	p_xml->QueryFloatAttribute("defaultShapeRestitution", &defaultShapeRestitution);

	// Standardreibung lesen
	float defaultShapeFriction = 0.0f;
	p_xml->QueryFloatAttribute("defaultShapeFriction", &defaultShapeFriction);

	// Tiles lesen
	for(const TiXmlElement* p_tile = p_xml->FirstChildElement("Tile");
		p_tile;
		p_tile = p_tile->NextSiblingElement("Tile"))
	{
		// Tile laden und eintragen
		Tile* p_theTile = new Tile(p_tile, defaultTileSize, defaultShapeRestitution, defaultShapeFriction);
		tiles[p_theTile->getID()] = p_theTile;
	}
}

TileSet::~TileSet()
{
	// alle Tiles lschen
	for(TileMapType::const_iterator it = tiles.begin();
		it != tiles.end();
		++it)
	{
		delete it->second;
	}
}

const Tile* TileSet::getTileByID(int id) const
{
	TileMapType::const_iterator it = tiles.find(id);
	return it == tiles.end() ? 0 : it->second;
}

void TileSet::drawTiles(const Uint32* p_tiles,
						const Vec2i& size,
						int pitch,
						const RGBA& color) const
{
	app.setTexturing(true);
	image->bind();

	Vec2f cursor(0.0f, 0.0f);

	glBegin(GL_QUADS);
	glColor4fv(color);

	const Uint32* p_row = p_tiles;
	for(int y = 0; y < size.y; ++y)
	{
		for(int x = 0; x < size.x; ++x)
		{
			// Tile zeichnen, wenn vorhanden
			if(p_row[x])
			{
				const Tile& tile = *tiles.find(p_row[x])->second;
				if(tile.isVisible()) tile.draw(cursor, color);
			}

			cursor.x += 1.0f;
		}

		// nchste Zeile
		p_row += pitch;
		cursor.x = 0.0f;
		cursor.y += 1.0f;
	}

	glEnd();
}
