/*****************************************************************/
/*      ____  ____  __    __    __  _______  ____  ___    ____   */
/*     / __ \/ __ \/ /   / /   /  |/  / __ \/ __ \/   |  / __ \  */
/*    / /_/ / / / / /   / /   / /|_/ / / / / /_/ / /| | / / / /  */
/*   / _, _/ /_/ / /___/ /___/ /  / / /_/ / _, _/ ___ |/ /_/ /   */
/*  /_/ |_|\____/_____/_____/_/  /_/\____/_/ |_/_/  |_/_____/    */
/*          ___ _   _ _  _   _   ___ ___   ___ __  _  __         */
/*         / __| | | | || | /_\ / __| __| |_  )  \/ |/  \        */
/*        | (_ | |_| | __ |/ _ \\__ \ _|   / / () | | () |       */
/*         \___|\___/|_||_/_/ \_\___/___| /___\__/|_|\__/        */
/*                                                               */
/*                    --- EIN TOLLES SPIEL ---                   */
/*                                                               */
/*         Jan Hapke, Wilhelm Mailnder, David Scherfgen         */
/*                                                               */
/*****************************************************************/

#include "pch.h"

const float M_PI_FLOAT = 3.1415927f;
const double M_PI_DOUBLE = 3.1415926535897932;
const float M_2_PI_FLOAT = 6.2831853f;
const double M_2_PI_DOUBLE = 6.2831853071795865;

SDL_Surface* downscaleSurface2x(SDL_Surface* p_surface,
								DownscaleFilter filter)
{
	assert(p_surface);
	assert(p_surface->format->BitsPerPixel == 32);
	assert(!(p_surface->w & 1) && !(p_surface->h & 1));

	// Surface mit der halben Auflsung erstellen
	SDL_Surface* p_result = SDL_CreateRGBSurface(p_surface->flags, p_surface->w / 2, p_surface->h / 2, p_surface->format->BitsPerPixel,
												 p_surface->format->Rmask, p_surface->format->Gmask, p_surface->format->Bmask, p_surface->format->Amask);
	assert(p_result);

	// Surfaces sperren
	SDL_LockSurface(p_surface);
	SDL_LockSurface(p_result);

	for(int y = 0; y < p_result->h; ++y)
	{
		for(int x = 0; x < p_result->w; ++x)
		{
			// die 4 umgebenden Pixel aus der Original-Surface holen
			const Uint32* p_original00 = reinterpret_cast<const Uint32*>(reinterpret_cast<const char*>(p_surface->pixels) + y * 2 * p_surface->pitch + x * 8);
			const Uint32* p_original10 = p_original00 + 1;
			const Uint32* p_original01 = reinterpret_cast<const Uint32*>(reinterpret_cast<const char*>(p_surface->pixels) + (y * 2 + 1) * p_surface->pitch + x * 8);
			const Uint32* p_original11 = p_original01 + 1;
			Uint32* p_out = reinterpret_cast<Uint32*>(reinterpret_cast<char*>(p_result->pixels) + y * p_result->pitch + x * 4);
			const RGBA original00(RGBA::fromRGBA32(*p_original00));
			const RGBA original10(RGBA::fromRGBA32(*p_original10));
			const RGBA original01(RGBA::fromRGBA32(*p_original01));
			const RGBA original11(RGBA::fromRGBA32(*p_original11));

			RGBA out(0.0f, 0.0f);
			switch(filter)
			{
			case DF_BOX:
				// alle sichtbaren Pixel gleich wichten
				if(original00.a) out += original00;
				if(original10.a) out += original10;
				if(original01.a) out += original01;
				if(original11.a) out += original11;
				out *= 0.25f;
				break;

			case DF_BOX_ALPHA_WEIGHTED:
				// die 4 Pixel mit ihrem Alphawert wichten
				// (sehr gut fr Schriftarten geeignet)
				out += original00.a * original00;
				out += original10.a * original10;
				out += original01.a * original10;
				out += original11.a * original11;
				out /= original00.a + original10.a + original01.a + original11.a;
				break;
			}

			// in die neue Surface schreiben
			*p_out = out.toRGBA32();
		}
	}

	// Surfaces entsperren
	SDL_UnlockSurface(p_surface);
	SDL_UnlockSurface(p_result);

	return p_result;
}

b2ShapeDef* transformShapeDef(const b2ShapeDef& shapeDef,
							  const Vec2f& translation,
							  const Vec2f& scaling)
{
	b2ShapeDef* p_result = 0;

	// Shape kopieren und verschieben
	switch(shapeDef.type)
	{
	case e_circleShape:
		{
			b2CircleDef* p_circleDef = new b2CircleDef(static_cast<const b2CircleDef&>(shapeDef));
			p_circleDef->localPosition += translation;
			assert(scaling.x == scaling.y);
			p_circleDef->radius *= scaling.x;
			p_result = p_circleDef;
		}
		break;
	case e_polygonShape:
		{
			b2PolygonDef* p_polygonDef = new b2PolygonDef(static_cast<const b2PolygonDef&>(shapeDef));
			for(int i = 0; i < p_polygonDef->vertexCount; ++i)
			{
				p_polygonDef->vertices[i] = translation + scaling * Vec2f(p_polygonDef->vertices[i]);
			}

			p_result = p_polygonDef;
		}
		break;
	}

	return p_result;
}

const std::string formatTime(float t)
{
	const int hseconds = static_cast<int>(t * 100.0f);
	const int seconds = hseconds / 100;
	const int minutes = seconds / 60;

	char buffer[32] = "";
	sprintf(buffer, "%02d:%02d.%02d", minutes, seconds % 60, hseconds % 100);

	return buffer;
}

const std::string strip(const std::string& s)
{
	for(std::string::const_iterator it = s.begin();
		it != s.end();
		++it)
	{
		if(!iswspace(*it)) return std::string(it, s.end());
	}

	return "";
}

const std::string toLower(const std::string& s)
{
	std::string result;
	result.reserve(s.length());

	for(std::string::const_iterator it = s.begin();
		it != s.end();
		++it)
	{
		result.push_back(isupper(*it) ? tolower(*it) : *it);
	}

	return result;
}

unsigned int nextPow2(unsigned int x)
{
	for(unsigned int i = 1; i < (1 << 31); i <<= 1) if(i >= x) return i;
	return x;
}

float easeInEaseOut(float x)
{
	return x * x * (3.0f - 2.0f * x);
}
